/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Bin
#else
#define __SEG__ Bin
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
/*
  local function to count the number of values n x in bins specified
  by bin boundaries bnds[0], bnds[1], ..., bnds[abs(nbins)]
  If nbins == 0, boundaries are assumed equal spaced with values
  bnds[0], bnds[0] + width, bnds[0] + 2*width, bnds[0] + 3*width, ..., where
  width = bnds[1] - bnds[0].
*/
#define BIGNBINS    8 /* threshold for quartering */

static void doBinning(double *x, double *bnds, double *counts,
					  long n, long nbins, long equispaced, long outs[3])
{
	long       i, jstart = 1, jbin, j1, j2, j3;
	long       nmissing = 0, bigNbins = (nbins > BIGNBINS);
	double     q1, q2, q3;
	double     b0 = bnds[0], bn = bnds[nbins], width, value;
	WHERE("doBinning");
	
	outs[0] = outs[1] = outs[2] = 0;
	
	if(equispaced)
	{
		width = bnds[1] - bnds[0];
	} /*if(equispaced)*/
	else if(bigNbins)
	{
		j1 = nbins/4;
		j2 = nbins/2;
		j3 = nbins - j1;
		q1 = bnds[j1++];
		q2 = bnds[j2++];
		q3 = bnds[j3++];
	}
	for(i=0;i<n;i++)
	{
		value = *x++;
		if(isMissing(value))
		{
			outs[2]++;
		}
		else if(value <= b0)
		{
			outs[0]++;
		}
		else if(bn < value)
		{
			outs[1]++;
		}
		else
		{
			if(equispaced)
			{
				jbin = ceil((value - b0)/width);
			}
			else
			{
				if(bigNbins)
				{
					jstart = (value <= q1) ? 1 :
							((value <= q2) ? j1 :
							 ((value <= q3) ? j2 : j3));
				}
				for(jbin = jstart; jbin < nbins ; jbin++)
				{
					if(value <= bnds[jbin])
					{
						break;
					}
				}
			}
			counts[jbin-1]++;
		}
	} /*for(i=0;i<n;i++)*/
	if(nmissing == n)
	{
		setMissing(counts[0]);
	}
} /*static void doBinning()*/

/*
  usage:
	bin(x)
	bin(x,nbins)   nbins integer > 0
	bin(x,cat(bndry,width))  bndry typical , width > 0
	bin(x,cat(b_0,b_1,b_2,...,b_nbins)) (b_0<b_1<...<b_nbins [,silent:T])
  Result: structure with components 'boundaries', 'counts'
	counts has nbins rows.  Values < b_0 or >= b_nbins are not
	counted
*/

Symbolhandle bin(Symbolhandle list)
{
	Symbolhandle          arg1 = COMPVALUE(list,0), arg2, arg3;
	Symbolhandle          result = (Symbolhandle) 0;
	Symbolhandle          boundaries = (Symbolhandle) 0;
	Symbolhandle          counts = (Symbolhandle) 0;
	long                  nargs = NARGS(list);
	long                  dim[2], nrows, ncols, nvalues, outs[3];
	long                  nbins, length = 0, i, j, equispaced = 1;
	long                  silent = 0;
	double                boundary0, bottom, top, binwidth, temp;
	double                extremes[2], range;
	char                 *keyword;
	WHERE("bin");
	
	setMissing(binwidth);
	if(nargs == 1 && arg1== (Symbolhandle) 0)
	{
		badNargs(FUNCNAME, -1001);
		goto errorExit;
	}
	if(nargs > 3)
	{
		badNargs(FUNCNAME, -3);
		goto errorExit;
	}
	if(!argOK(arg1,0,(nargs == 1) ? 0 : 1))
	{
		goto errorExit;
	}

	if(TYPE(arg1) != REAL || !isMatrix(arg1,dim))
	{
		sprintf(OUTSTR,
				"ERROR: argument% to %s not REAL matrix",
				(nargs == 1) ? "" : " 1");
		goto errorExit;
	}
	nrows = dim[0];
	ncols = dim[1];
	
	symExtremes(arg1,extremes);
	if(extremes[1] == HUGE)
	{
		sprintf(OUTSTR,
				"ERROR: all values of first argument to %s are MISSING",
				FUNCNAME);
		goto errorExit;
	}
	
	range = extremes[1] - extremes[0];
	if(range <= 0.0)
	{
		range = 1;
	}
	
	if(nargs == 1)
	{
		nbins = ceil(log((double) nrows) / log(2.0) + 1);
	}
	else
	{
		arg2 = COMPVALUE(list,1);
		if(!argOK(arg2,REAL,2))
		{
			goto errorExit;
		}
		length = symbolSize(arg2);

		if(!isVector(arg2))
		{
			sprintf(OUTSTR,
					"ERROR: argument 2 to %s must be REAL scalar or vector",
					FUNCNAME);
		}
		else if(anyMissing(arg2))
		{
			sprintf(OUTSTR,
					"ERROR: argument 2 to %s contains MISSING values ",
					FUNCNAME);
		}
		else if(isScalar(arg2))
		{
			nbins = DATAVALUE(arg2,0);
			if(nbins <= 0 || DATAVALUE(arg2,0) != (double) nbins)
			{
				sprintf(OUTSTR,
						"ERROR: number of bins must be positive integer for %s()",
						FUNCNAME);
			}
		}
		else if(length == 2)
		{
			boundary0 = DATAVALUE(arg2,0);
			binwidth = DATAVALUE(arg2,1);
			if(binwidth <= 0.0)
			{
				sprintf(OUTSTR,
						"ERROR: bin width must be positive for %s()", FUNCNAME);
			}
			else 
			{
				boundary0 -= floor((boundary0 - extremes[0])/binwidth) *
					binwidth;
				if(boundary0 >= extremes[0])
				{
					boundary0 -= binwidth;
				}
				nbins = ceil((extremes[1] - boundary0)/binwidth);
			}			
		}
		else
		{
			nbins = length - 1;
			boundary0 = bottom = DATAVALUE(arg2,0);
			binwidth = DATAVALUE(arg2,1) - bottom;
			for(i=1;i<=nbins;i++)
			{
				top = DATAVALUE(arg2,i);
				temp = top - bottom;
				if(temp <= 0.0)
				{
					sprintf(OUTSTR,
							"ERROR: bin boundaries must be strictly increasing");
					goto errorExit;
				}
				if(equispaced && fabs(temp - binwidth)/binwidth > 1e-8)
				{
					equispaced = 0;
				}
				bottom = top;
			}
		}
		if(nargs == 3)
		{
			if(!argOK((arg3 = COMPVALUE(list,2)),0,3))
			{
				goto errorExit;
			}
			if((keyword =isKeyword(arg3)))
			{
				if(strcmp(keyword,"silent") != 0)
				{
					badKeyword(FUNCNAME,keyword);
					goto errorExit;
				}
			}
			else
			{
				keyword = "silent";
			}
			if(!isScalar(arg3) || TYPE(arg3) != LOGIC)
			{
				notTorF(keyword);
				goto errorExit;
			}
			silent = (DATAVALUE(arg3,0) != 0.0);
		}
	} /*if(nargs == 1){...}else{...}*/		
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	if(isMissing(binwidth))
	{
		boundary0 = extremes[0] - .025*range;
		binwidth = 1.1 * range / (double) nbins;
	}
	
	result = StrucInstall(SCRATCH, 2);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	nvalues = nbins * ncols;
	counts = COMPVALUE(result,1) = Makereal(nvalues);
	if(counts == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	if(ncols > 1)
	{
		setNDIMS(counts,2);
		setDIM(counts,1,nbins);
		setDIM(counts,2,ncols);
	}
	setNAME(counts,"counts");
	
	symFill(counts,0.0);
	
	if(length > 2 && isscratch(NAME(arg2)))
	{ /* reuse arg2 for boundaries */
		COMPVALUE(list,1) = (Symbolhandle) 0;
		Cutsymbol(arg2);
		boundaries = COMPVALUE(result,0) = arg2;
	}
	else
	{
		boundaries = COMPVALUE(result,0) = Makereal(nbins+1);
		if(boundaries == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		if(length > 2)
		{
			for(i=0;i<length;i++)
			{
				DATAVALUE(boundaries,i) = DATAVALUE(arg2,i);
			}
		}
		else
		{
			for(i=0;i<=nbins;i++)
			{
				DATAVALUE(boundaries,i) = boundary0 + i*binwidth;
			}
		}
	}
	setNAME(boundaries,"boundaries");
	setNCLASS(boundaries, -1);

	for(j=0;j<ncols;j++)
	{
		doBinning(DATAPTR(arg1) + j * nrows, DATAPTR(boundaries),
				  DATAPTR(counts) + j*nbins, nrows, nbins, equispaced,
				  outs);
		if(!silent)
		{
			if(outs[0] + outs[1] > 0)
			{
				sprintf(OUTSTR,
						"WARNING: %ld low and %ld high values in column %ld not counted by %s",
						outs[0], outs[1], j+1,FUNCNAME);
				putOUTSTR();
			}
			if(outs[2] == nrows)
			{
				sprintf(OUTSTR,
						"WARNING: Column %ld of data for %s is all MISSING",
						j+1, FUNCNAME);
			}
			else if(outs[2] > 0)
			{
				sprintf(OUTSTR,
						"WARNING: %ld MISSING values in column %ld of data for %s",
						outs[2], j+1, FUNCNAME);
			}
			putOUTSTR();
		} /*if(!silent)*/
	} /*for(j=0;j<ncols;j++)*/
	
	return (result);
	
  errorExit:
	putOUTSTR();
	
	Removesymbol(result);
	
	return (0);
}
