/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Funbalan
#else
#define __SEG__ Funbalan
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>
#include "dbug.h"

static long            getcellindex(long term[], long k)
{
	long            indices[MAXVARS], i, j;
	long            nvars = (long) NVARS;
	
	for (i = 0; i < nvars; i++)
	{
		indices[i] = k % NCLASSES[i];
		k = k / NCLASSES[i];
	}
	j = 1;
	k = 0;
	for (i = 0; i < nvars; i++)
	{
		if (term[i] > 0)
		{
			k += j * indices[i];
			j *= NCLASSES[i];
		}
	}
	return (k);
}

/*
	Fast iterative routine for computing ss and df for unbalanced designs
	Uses EM algorithm
	
	
	RESIDUALS	handle to response vector (which will be changed)
	Y       handle to response variable (will be changed)
	X		array of handles to class variables
	NVARS	numberof variables (columns of x)
	NCLASSES	number of classes in each variable
	NDATA	number of data values (will be changed)
	MODEL	handle giving terms in order to be used in model
		term i contains the variables in the binary expansion of i
		0 is constant; 3 is AB; 5 is AC etc
	SS	handle to term sums of squares
	DF	handle of term df
*/

#define GCTS   0
#define GMNS   1
#define GRESID 2
#define NTRASH 3

long funbalanova(maxiter,epsilon)
long                maxiter;
double              epsilon;
{
	Symbolhandle    symh;
	long            i, j, k, m, n, i2, termmax, ncells, npseudo, term[MAXVARS];
	long            maxcts, df,nEmpty = 0;
	long            nvars = (long) NVARS, ndata = (long) NDATA, ndataOrig;
	long            nterms = (long) NTERMS, notmissing = (long) NOTMISSING;
	long            returnValue = 0;
	modelPointer    iterm;
	double        **cts = (double **) 0, **mns = (double **) 0;
	double        **resid = (double **) 0;
	double        **vars = (double **) 0; /* used only as dummy argument */
	double          curss, newss, oldss, ybar;
	WHERE("funbalanova");
	TRASH(NTRASH,errorExit); /* room for up to 3 scratch */
	
	SETUPINT(errorExit);

	ncells = 1;
	for (i = 0; i < nvars; i++)
	{
		ncells *= NCLASSES[i];
	}

	if(!getScratch(cts,GCTS,ncells,double) ||
	   !getScratch(mns,GMNS,ncells,double))
	{
		goto errorExit;
	}

	/* compute df's */

	for (i = 0; i < nterms; i++)
	{
		iterm = modelterm(MODEL,i);	/* terms in model */
		(*DF)[i] = 1;
		for (k = 0; k < nvars; k++)
		{
			if(inTerm(k,iterm))
			{
				(*DF)[i] *= (double) NCLASSES[k];
			}
		}/*for (k = 0; k < nvars; k++)*/
	} /*for (i = 0; i < nterms; i++)*/

	/* DF now has maximum possible df, go through and take out of higher terms */

	for (i = 0; i < nterms - 1; i++)
	{
		iterm = modelterm(MODEL,i);
		for (j = i + 1; j < nterms; j++)
		{
			if(inEffectiveTerm(modelterm(MODEL,j),iterm))
			{	/* iterm included in in term j */
				(*DF)[j] -= (*DF)[i];
			}
		}
	}

	
	/* at this point, DF has maximum possible, but there may be missing cells
	   now go through and attempt to reduce DF for missing cells.
	   The current algorithm does not work properly in certain cases, e.g.
	   the following cell pattern (O = Empty, X = not empty)
	       X   X   O   O   O
	       X   X   O   O   O
	       O   O   X   X   X
	       O   O   X   X   X
	       O   O   X   X   X
	*/
	
	/*
	  make a structure for used cells, cts tells us what has been
	  used up till now, and mns will be for the current model term
	*/

	/* compute cell counts into (*cts) */
	for (i = 0; i < nvars; i++)
	{
		term[i] = 1;
	}
	(void) cellstat(Y, term, (double **) 0, (double **) 0, cts, CELLCOUNTS);

	for (k = 0; k < ncells; k++)
	{ /* count empty cells */
		if ((*cts)[k] == 0)
		{
			nEmpty++;
		}
	}
	if(nEmpty > 0 && PRINTWARNINGS)
	{
		sprintf(OUTSTR,
				"WARNING: There %s %ld empty cell%s; coefs() may give wrong answers",
				(nEmpty == 1) ? "is" : "are", nEmpty,
				(nEmpty == 1) ? "" : "s");
		putOUTSTR();
		myerrorout("         and the degrees of freedom may be in error");
	} /*if(nEmpty > 0 && PRINTWARNINGS)*/

	for (i = 0; i < nterms; i++)
	{
		df = (long) (*DF)[i];
		/* set term to binary expansion of i-th term */
		getterm(modelterm(MODEL,i), term, nvars);
		/* set up cell counter for term */
		termmax = 1;
		for (j = 0; j < nvars; j++)
		{
			if (term[j] > 0)
			{
				termmax *= NCLASSES[j];
			}
		}
		/* termmax is size of associated marginal table */
		for (j = 0; j < termmax; j++)
		{
			(*mns)[j] = 0.;
		}
		/*
		  go through all  previously used cells and mark locations
		  in current term counter
		*/
		for (k = 0; k < ncells; k++)
		{
			if ((*cts)[k] < 0)
			{
				(*mns)[getcellindex(term, k)] = -1.;
			}
		}
		/*
		  go through all "unused" cells and see if they can be used in
		  new term.  Cumulate the number added in
		  (*mns)[m] is set to -1 if the m-th marginal cell is not empty
		  It is set to k+1 if the cell is empty, k = index of first cell of
		  table associated with m-th marginal cell.  Cell k will be marked as
		  "used up" if m-th cell is empty.
		*/
		for (k = 0; k < ncells; k++)
		{
			if(((*mns)[m = getcellindex(term, k)] >= 0.))
			{
				if ((*cts)[k] > 0)
				{
					if((*mns)[m] == 0.0)
					{
						(*cts)[k] = -1.;
					}
					(*mns)[m] = -1;
				}
				else if((*cts)[k] == 0.0 && (*mns)[m] == 0.0)
				{ /* mark this level as including empty cell */
					(*mns)[m] = (double) (k + 1);
				}
			}
		}

		for (j = 0; j < termmax; j++)
		{
			if((m = (long) (*mns)[j]) > 0)
			{
				df--; /* deduct d.f. */
				(*cts)[m - 1] = -1.0; /* mark cell as used */
			}
		}
		if(df < 0)
		{
			df = 0;
		}
		
		(*DF)[i] = df;
	} /* for i */

	(*DF)[nterms] = NOTMISSING;
	for (i = 0; i < nterms; i++)
	{
		(*DF)[nterms] -= (*DF)[i];
	}

	/* find out how many pseudo responses are needed and set them up */

	/* recompute cell counts into (*cts) */
	for (i = 0; i < nvars; i++)
	{
		term[i] = 1;
	}
	(void) cellstat(Y, term, (double **) 0, (double **) 0, cts, CELLCOUNTS);

	maxcts = 0; /* max((*cts)[i],i=1,ncells) */
	for (i = 0; i < ncells; i++)
	{
		if ((*cts)[i] > maxcts)
		{
			maxcts = (*cts)[i];
		}
	}
	ndataOrig = ndata;
	ndata = maxcts * ncells;
	NDATA = (double) ndata;
	npseudo = ndata - notmissing;

	/*  squeeze out cases with missing data  */
	k = 0;
	if(notmissing != ndataOrig)
	{
		for (i = 0; i < ndataOrig; i++)
		{
			if (!isMissing((*Y)[i]))
			{
				for (j = 0; j < nvars; j++)
				{
					if (isMissing((*(X[j]))[i]))
					{
						break;
					}
				}
				if (j == nvars)
				{ /* no missing X values */
					(*Y)[k] = (*Y)[i];
					for (j = 0; j < nvars; j++)
					{
						(*(X[j]))[k] = (*(X[j]))[i];
					}
					k++;
				} /*if (j == nvars)*/
			} /*if (!isMissing((*Y)[i]))*/
		} /* for (i = 0; i < ndataOrig; i++) */
	} /*if(notmissing != ndataOrig)*/

	ybar = 0.0;
	for(i=0;i<notmissing;i++)
	{
		ybar += (*Y)[i];
	}
	ybar /= NOTMISSING;

	if(ndata != ndataOrig)
	{ /* change length of variables */
		symh = MODELVARS[0];
		
		/* increase the length of variables as if it were balanced */
		Y = (double **) mygrowhandle((char **) Y,ndata * sizeof(double));

		/* Y should always be identical with DATA(MODELVARS[0]) */
		setDATA(symh,Y);
		if(Y == (double **) 0)
		{
			goto errorExit;
		}
		setDIM(symh,1,ndata);

		for (i = 0; i < nvars; i++)
		{
			symh = MODELVARS[i+1];
			X[i] = (double **) mygrowhandle((char **) X[i],
											ndata * sizeof(double));
			/* X[i] should always be identical with DATA(MODELVARS[i+1]) */
			setDATA(symh,X[i]);
			if(X[i] == (double **) 0)
			{
				goto errorExit;
			}
			setDIM(symh,1,ndata);
		}
	} /*if(ndata != ndataOrig)*/

	if(npseudo != 0)
	{/* add pseudo responses, if any, at end */
		k = notmissing;
		for (i = 0; i < nvars; i++)
		{ /* start in cell (1,1,...,1) */
			term[i] = 1;
		}
		for (i = 0; i < ncells; i++)
		{
			i2 = maxcts - (*cts)[i];
			for (n = 0; n < i2; n++)
			{
				(*Y)[k] = ybar;
				for (m = 0; m < nvars; m++)
				{
					(*(X[m]))[k] = term[m];
				}
				k++;
			}
			/* step factor levels */
			term[0]++;
			for (j = 1; j < nvars; j++)
			{
				if (term[j - 1] > NCLASSES[j - 1])
				{
					term[j - 1] = 1;
					term[j]++;
				}
			}
		} /*for (i = 0; i < ncells; i++)*/
	} /*if(npseudo != 0)*/

	/* now go through terms in model getting residual ss for each term
	   and sweeping out effects */

	curss = 0.0;
	for (i = 0; i < notmissing; i++)
	{
		curss += (*Y)[i] * (*Y)[i];	/*  get total SS */
	}
	
	if(!getScratch(resid,GRESID,ndata,double))
	{
		goto errorExit;
	}
	
	for (i = 0; i < nterms; i++)
	{
		oldss = 0.0;
		for (j = 0; j < ndata; j++)
		{
			(*resid)[j] = (*Y)[j];
		}
		for (k = 0; k < maxiter; k++)
		{
			if (interrupted() != INTNOTSET)
			{
				goto errorExit;
			}
			for (j = 0; j <= i; j++)
			{
				/* set binary expansion in term */
				getterm(modelterm(MODEL,j), term, nvars);
				/* sweep cell means from current residuals, compute new RSS */
				newss = cellstat(resid, term, mns, vars, cts,
								 CELLMEANSANDSWEEP);
			}
			if (fabs(oldss - newss) <= epsilon * (oldss + newss))
			{
				break;
			}
			oldss = newss;
			if (k < maxiter - 1)
			{
				for (j = 0; j < ndata - npseudo; j++)
				{
					(*resid)[j] = (*Y)[j];
				}
				/* set pseudo obs to the current fitted values */
				for(j = ndata - npseudo;j < ndata;j++)
				{
					(*resid)[j] = (*Y)[j] - (*resid)[j];
					(*Y)[j] = (*resid)[j];
				}
			}
		} /*for (k = 0; k < maxiter; k++)*/

		/* set pseudo obs to the final fitted values */
		for(j = ndata - npseudo;j < ndata;j++)
		{
			(*Y)[j] = (*Y)[j] - (*resid)[j];
		}
		if(k == maxiter)
		{
			returnValue = 1;
		}
		
		(*SS)[i] = curss - newss;	/* set ss for term */
		curss = newss;
	}

	(*SS)[nterms] = curss;	/* error ss */

	unTrash(GRESID); /* resid is a keeper */
	RESIDUALS = resid;
	returnValue = (returnValue != 0) ? returnValue : 2;
	/* fall through */

  errorExit:
	*OUTSTR = '\0';
	emptyTrash();	/* dispose of scratch space */

	RETURN (returnValue);
	
}

