/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef MPW
#ifdef MPW3
#pragma segment Glm
#else
#define __SEG__ Glm
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "blas.h"
#include "dbug.h"

/* 
  Front end for functions 
	anova(model [,wts:vector])
	fastanova(model)
	ipf(model [,inc:T])
	logistic(model, denom [,offset:vector] [,inc:T])
	manova(model [,wts:vector])
	poisson(model [,offset:vector] [,inc:T])
	regress(model [,wts:vector])
	robust(mode, [inc:T])
	screen(model[,keep:charVec,print:mbest:ival,forced:vector,
		method:string,s2:posval,penalty:posval])
	wtanova(model,wts)
	wtmanova(model,wts)
	wtregress(model,wts)
*/
typedef struct opEntry
{
	char *name;
	short iop;
	short nargs;
} opEntry;

static opEntry Ops[] =
{ /* name        iop             nargs           arguments & keywords     */
	"anova",     BALANOVA,        -5, /*model,print|silent,wts,coefs,unbal*/
	"fastanova", FASTANOVA,       -4, /*model,print|silent,maxiter,eps*/
	"ipf",       IPF,             -5, /*model,print|silent,maxiter,eps,inc*/
	"logistic",  LOGITREG,        -7, /*model,n,print|silent,offset,maxiter,
                                        eps,inc,coefs*/
	"manova",    MANOVA,          -4, /*model,print|silent,wts,coefs*/
	"poisson",   POISSONREG,      -7, /*model,print|silent,offset,maxiter,
                                        eps,inc,coefs*/
	"regress",   OLSREG,          -3, /*model,print|silent,wts*/
	"robust",    ROBUSTREG,       -5, /*model,print|silent,maxiter,eps,inc*/
	"screen",    LEAPS,           -8, /*model,print,keep,mbest,force,method,
                                        s2|ssq,penalty*/
	"wtanova",   UNBALANO | CASEWEIGHTS, -4, /*model,wts,print|silent,coefs*/
	"wtmanova",  MANOVA | CASEWEIGHTS,   -4, /*model,wts,print|silent,coefs*/
	"wtregress", OLSREG | CASEWEIGHTS,   -3, /*model,wts,print|silent*/
	(char *) 0,  0,                0
};

static long            FactorFound = 0;

/* default constants*/
#define FASTANOVAMAXITER  25
#define FASTANOVAEPSILON  1e-6
#define ITERGLMMAXITER    50
#define ITERGLMEPSILON    1e-6
#define IPFMAXITER        25
#define IPFEPSILON        1e-6


Symbolhandle    glm(list)
Symbolhandle    list;
{

	/* control  routine for glms */

	Symbolhandle    symh, symh1, symhKey;
	Symbolhandle    symWRES, symSS, symDF, symHII, symXTX, symCOEF, symRES;
	char           *keyword = (char *) 0;
	char          **tmpSTRMODEL = (char **) 0;
	long            i, j;
	long            incr = 1, length;
	long            nargs = NARGS(list), margs;
	long            nvars, ndata, nregx, ny, nterms;
	long            mdatay, mdatax, mdatass, mdataxx;
	long            printResults = 1, usePrevious = 0, silent = 0;
	long            docoefs = DOCOEFS;
	long            prevType = 0, notInstalled = 0;
	long            maxiter = -1;
	double          epsilon;
	long            startKey = 1;
	long            keyindex;
	int             reply = 2, doCleanup = 0;
	opEntry        *op;
	double          val;
	double        **tmp;
	WHERE("glm");
	TRASH(1,errorExit); /* room for 1 scratch handle */

	SETUPINT(errorExit);
	
	MODELTYPE = NOMODEL;
	setMissing(epsilon);
	*OUTSTR = '\0';
	
	/* determine model type from FUNCNAME */
	for(op = Ops; op->iop != 0; op++)
	{
		if(strcmp(FUNCNAME,op->name) == 0)
		{
			MODELTYPE = op->iop;
			margs = op->nargs;
			break;
		}
	}

	if(nargs > labs(margs))
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}

	if((keyindex = findKeyword(list,"print",startKey)) >= 0 ||
	   (keyindex = findKeyword(list,"silent",startKey)) >= 0 )
	{
		symhKey = COMPVALUE(list,keyindex);
		keyword = isKeyword(symhKey);
		if(MODELTYPE & LEAPS)
		{
			if (*keyword == 's')
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
			/* screen() decodes print*/
		} /*if(MODELTYPE & LEAPS)*/
		else 
		{
			if(!isScalar(symhKey) || TYPE(symhKey) != LOGIC)
			{
				notTorF(keyword);
				goto errorExit;
			}
		
			if(*keyword == 'p')
			{
				printResults = (DATAVALUE(symhKey,0) != 0.0);
			}
			else
			{
				silent = (DATAVALUE(symhKey,0) != 0.0);
				printResults = !silent;
			}
			setNAME(symhKey,USEDKEYWORD);
			/* rotate it to the end of list */
			if(keyindex != nargs-1)
			{
				for(i=keyindex;i<nargs-1;i++)
				{
					COMPVALUE(list, i) = COMPVALUE(list, i+1);
				}
				COMPVALUE(list, nargs-1) = symhKey;
			}
		} /*if(MODELTYPE & LEAPS){}else{}*/
	}

	if((keyindex = findKeyword(list,"coef",startKey)) >= 0)
	{
		if(MODELTYPE & LEAPS)
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		symhKey = COMPVALUE(list,keyindex);
		keyword = isKeyword(symhKey);

		if(!isScalar(symhKey) || TYPE(symhKey) != LOGIC)
		{
			notTorF(keyword);
			goto errorExit;
		}

		setNAME(symhKey,USEDKEYWORD);
		docoefs = (DATAVALUE(symhKey, 0) != 0) ? DOCOEFS : DONTDOCOEFS;
		
		if(docoefs == DONTDOCOEFS &&
		   (MODELTYPE & OLSREG || MODELTYPE & FASTANOVA ||
			MODELTYPE & IPF || MODELTYPE == ROBUSTREG))
		{
			sprintf(OUTSTR,
					"ERROR: %s:F is illegal on %s()", keyword, FUNCNAME);
			goto errorExit;
		}

		/* rotate it to the end of list */
		if(keyindex != nargs-1)
		{
			for(i=keyindex;i<nargs-1;i++)
			{
				COMPVALUE(list, i) = COMPVALUE(list, i+1);
			}
			COMPVALUE(list, nargs-1) = symhKey;
		}
	} /*if((keyindex = findKeyword(list,"coef",startKey)) >= 0)*/

	if((MODELTYPE & BALANOVA || MODELTYPE & MANOVA || MODELTYPE & OLSREG))
	{ /* allow weights in anova(), manova(), and regress() */
		if((keyindex = findKeyword(list,"wt",startKey)) >= 0 ||
		   (keyindex = findKeyword(list,"weight",startKey)) >= 0)
		{
			symhKey = COMPVALUE(list, keyindex);
			/* type will be checked by getmodel() */
			if(MODELTYPE & BALANOVA)
			{
				MODELTYPE = UNBALANO;
			}
			MODELTYPE |= CASEWEIGHTS;
			setNAME(symhKey,USEDKEYWORD);
			 /* make sure it is in the right place */
			if(keyindex != 1)
			{ /* swap in list */
				COMPVALUE(list,keyindex) = COMPVALUE(list,1);
				COMPVALUE(list,1) = symhKey;
			}
			startKey = nargs = 2;
		}

		if((keyindex = findKeyword(list,"unbal",startKey)) >= 0)
		{
			symhKey = COMPVALUE(list,keyindex);
			setNAME(symhKey,USEDKEYWORD);
			if(TYPE(symhKey) != LOGIC || !isScalar(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if(MODELTYPE & BALANOVA)
			{
				if(DATAVALUE(symhKey,0) != 0)
				{
					MODELTYPE = UNBALANO;
				}
			}
			else if(!silent)
			{
				sprintf(OUTSTR,
						"WARNING: keyword %s ignored by %s",keyword,FUNCNAME);
				putOUTSTR();
			}
			nargs = 1;
		} /*if((keyindex = findKeyword(list,"unbal",1)) >= 0)*/
	} /*if((MODELTYPE & BALANOVA||MODELTYPE & MANOVA||MODELTYPE & OLSREG))*/
	
	if(MODELTYPE & BALANOVA)
	{
		docoefs = DOCOEFS;
	}
	MODELTYPE |= docoefs;
	
	if(MODELTYPE & (ITERWEIGHTS | FASTANOVA | IPF))
	{
		startKey = (MODELTYPE & LOGITREG) ? 2 : 1;

		for(i=startKey;i<nargs;i++)
		{
			symhKey = COMPVALUE(list,i);
			if((keyword = isKeyword(symhKey)) && 
			   (strncmp(keyword,"maxit",5) == 0 || 
				strncmp(keyword,"eps",3) == 0))
			{
				if(!argOK(symhKey,0,i+1))
				{
					goto errorExit;
				}
				if(TYPE(symhKey) != REAL || !isScalar(symhKey))
				{
					notNumberOrReal(keyword);
					goto errorExit;
				}
				
				if(*keyword == 'm')
				{ /* maxiter */
					val = DATAVALUE(symhKey,0);
					if(val <= 0.0 || val != floor(val))
					{
						sprintf(OUTSTR,
								"ERROR: %s must be positive integer",keyword);
						goto errorExit;
					}
					maxiter = (long) val;
				} /*if(*keyword == 'm')*/
				else
				{ /* epsilon */
					epsilon = DATAVALUE(symhKey,0);
					if(epsilon <= 0.0)
					{
						sprintf(OUTSTR,
								"ERROR: %s must be positive REAL",keyword);
						goto errorExit;
					}
				} /*if(*keyword == 'm'){...}else{...}*/
				setNAME(symhKey,USEDKEYWORD);
			}
		} /*for(i=startKey;i<nargs;i++)*/
	} /*if(MODELTYPE&ITERWEIGHTS || MODELTYPE&FASTANOVA || MODELTYPE&IPF)*/
			   
	/* get model */
	symh = COMPVALUE(list,0);
	symh1 = Lookup("STRMODEL");
	/* preserve PREVMODELTYPE */
	prevType = PREVMODELTYPE;
	if(symh == (Symbolhandle) 0)
	{ /* argument 1 missing; use model in symbol STRMODEL, if it exists */
		if(symh1 == (Symbolhandle) 0)
		{
			sprintf(OUTSTR,
					"ERROR: no model specified for %s and no previous model available",
					FUNCNAME);
			goto errorExit;
		}
	/* Note : PREVMODELTYPE is reset in Symbol.c if STRMODEL is assigned to*/

		/* do not recompute things in certain cases */
		if(nargs == 1 && NVARS >= 0 && SS != (double **) 0 && 
		   PREVMODELTYPE & OLSREG &&
		   (MODELTYPE & BALANOVA ||
			(MODELTYPE & UNBALANO && (PREVMODELTYPE & CASEWEIGHTS))))
		{
			usePrevious = 1;
			if(PREVMODELTYPE & CASEWEIGHTS)
			{
				MODELTYPE = (UNBALANO | CASEWEIGHTS | PREVMODELTYPE & DOCOEFS);
			}
		}
	}
	else if(TYPE(symh) != CHAR || !isScalar(symh))
	{
		sprintf(OUTSTR, "ERROR: model for %s must be a single CHAR string",
				FUNCNAME);
		goto errorExit;
	}

	if(symh == (Symbolhandle) 0)
	{ /* use STRMODEL as current model */
		symh = symh1;
	}
	
	doCleanup = 1;

	if(!getScratch(tmpSTRMODEL,0,strlen(STRINGPTR(symh)) + 1,char))
	{
		goto errorExit;
	}
	strcpy(*tmpSTRMODEL,STRINGPTR(symh));

	if(!usePrevious)
	{
		glmcleanup(0);
		doCleanup = 0;
	}
	else
	{
		mydisphandle(STRMODEL);
	}
	
	if ((STRMODEL = setStrmodel(tmpSTRMODEL,1)) == (char **) 0)
	{
		doCleanup = 0;
		goto errorExit;
	}

/* *tmpSTRMODEL and *STRMODEL now contain the model without and with '1+' */
	if(!silent && !(MODELTYPE & LEAPS))
	{
		myprint("Model used is ");
		myerrorout(*tmpSTRMODEL);
	}
	emptyTrash(); /* get rid of tmpSTRMODEL */
	
/* Install *STRMODEL as variable STRMODEL */
	Remove("STRMODEL");
	symh = Install("STRMODEL", CHAR);
	if (symh == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	TMPHANDLE = mygethandle((long) strlen(*STRMODEL) + 1);
	setSTRING(symh,TMPHANDLE);
	if (TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}
	strcpy(STRINGPTR(symh), *STRMODEL);
	setNDIMS(symh,1);
	setDIM(symh,1,1);

	if (!usePrevious && getmodel(list, nargs, silent) != 0)
	{
		goto errorExit;
	}

	ndata = (long) NDATA;
	ny = (long) NY;
	nvars = (long) NVARS;
	nterms = (long) NTERMS;

	mdatay = ndata * ny; /* number of elements in dependent variable(s) */
	mdatass = ny * ny * (nterms + 1); /* number of elements in SS */
	if (MODELTYPE & LEAPS)
	{
		/* leaps does special things */
		RETURN (screen(list));
	}

	for(i = 0;i<nargs;i++)
	{
		if((keyword = isKeyword(symh = COMPVALUE(list,i))) &&
		   strcmp(NAME(symh),USEDKEYWORD) != 0)
		{
			sprintf(OUTSTR,
					"ERROR: unrecognized or duplicate keyword %s in %s",
					keyword, FUNCNAME);
			goto errorExit;
		}
	}
	if(MODELTYPE & CASEWEIGHTS && !silent)
	{
		myerrorout("Model fit by weighted least squares");
	}
	
	if(USEGLMOFFSET && !silent)
	{
		myerrorout("Vector of offsets used");
	}
	
	if(!usePrevious)
	{
		/* for regression we ignore class structure */
		FactorFound = 0;
		for (i = 0; i < nvars; i++)
		{
			if(NCLASSES[i] > 0)
			{
				FactorFound++;
			}
			else if(MODELTYPE & FASTANOVA)
			{
				sprintf(OUTSTR,
						"ERROR: %s cannot be used with variates in model",
						FUNCNAME);
				goto errorExit;
			}

			if (MODELTYPE & OLSREG)
			{
				NCLASSES[i] = -1;
/*				setNCLASS(MODELVARS[i+1], -1);*/
			}
		}
		
		SS = (double **) mygethandle(mdatass * sizeof(double));
		DF = (double **) mygethandle((nterms + 1) * sizeof(double));
		if (SS == (double **) 0 || DF == (double **) 0)
		{
			goto errorExit;
		}
		if(!(MODELTYPE & FASTANOVA))
		{
			RESIDUALS = (double **) mygethandle(mdatay * sizeof(double));
			HII = (double **) mygethandle(ndata * sizeof(double));
			if (RESIDUALS == (double **) 0 || HII == (double **) 0)
			{
				goto errorExit;
			}
			/* set residuals to Y*/
			DCOPY(&mdatay, *Y, &incr, *RESIDUALS, &incr);
		}
		
		/* set up X matrix if needed */
		if (!(MODELTYPE & (BALANOVA | FASTANOVA | IPF)) && Xsetup() > 0)
		{
			goto errorExit;
		}
		if(MODELTYPE & (BALANOVA | FASTANOVA | IPF))
		{
			nregx = 0;
			for(i=0;i<nterms;i++)
			{
				nregx += buildXmatrix(i,(double *) 0, MODELINFO);
			}
			NREGX = (double) nregx;
		}
		else
		{
			nregx = (long) NREGX;
		}
		mdatax = ndata * nregx; /* number of elements in indep. variable(s) */
		mdataxx = nregx * nregx; /* number of elements in XTXINV */
		if (MODELTYPE & (ITERWEIGHTS | CASEWEIGHTS | IPF))
		{
			WTDRESIDUALS = (double **) mygethandle(mdatay * sizeof(double));

			if (WTDRESIDUALS == (double **) 0)
			{
				goto errorExit;
			}
			if(!(MODELTYPE & IPF) && !(MODELTYPE & ROBUSTREG))
			{
				REGX2 = (double **) mygethandle(mdatax * sizeof(double));
				if (REGX2 == (double **) 0)
				{
					goto errorExit;
				}
				DCOPY(&mdatax, *REGX, &incr, *REGX2, &incr);
			} /*if(!(MODELTYPE & IPF))*/
		} /*if (MODELTYPE & (ITERWEIGHTS | CASEWEIGHTS | IPF))*/

		if (MODELTYPE & (IPF | BALANOVA))
		{
			if(maxiter < 0)
			{
				maxiter = IPFMAXITER;
			}
			if(isMissing(epsilon))
			{
				epsilon = IPFEPSILON;
			}
			if ((reply = ipf(maxiter,epsilon)) < 1)
			{
				goto errorExit;
			}
		} /*if (MODELTYPE & (IPF | BALANOVA))*/
		else if (MODELTYPE & FASTANOVA)
		{
			if(maxiter < 0)
			{
				maxiter = FASTANOVAMAXITER;
			}
			if(isMissing(epsilon))
			{
				epsilon = FASTANOVAEPSILON;
			}
			
			/* 
			  NB: funbalanova changes NDATA and size of Y and RESIDUALS
			  These modified values are needed for predtable() and coef()
			  and should not be changed back.  However, ndata still reflects
			  the true length of the response vector.
			*/
			if ((reply = funbalanova(maxiter,epsilon)) < 1)
			{
				goto errorExit;
			}
		} /*else if (MODELTYPE & FASTANOVA)*/
		else if (MODELTYPE & ITERWEIGHTS)
		{
			if(maxiter < 0)
			{
				maxiter = ITERGLMMAXITER;
			}
			if(isMissing(epsilon))
			{
				epsilon = ITERGLMEPSILON;
			}
			if ((reply = iterglm(maxiter,epsilon,silent)) < 1)
			{
				goto errorExit;
			}
		}/*else if (MODELTYPE & ITERWEIGHTS)*/
		else
		{
			if (unbalanova() < 1)
			{
				goto errorExit;
			}
		}
	} /*if(!usePrevious)*/

	if(reply != 2)
	{
		sprintf(OUTSTR,
				"WARNING: %s failed to converge fully; results may be inaccurate",
				FUNCNAME);
		putOUTSTR();
	}
	
	if(usePrevious)
	{
		/*
		  PREVMODELTYPE is changed to current model unless we are reusing
		  results of previous regression model when it is unchanged
		*/
		PREVMODELTYPE = prevType;
	}

	while(!usePrevious) /* fake while to allow exiting {...} by break*/
	{
		/* set up user accessible symbols for various quantities */
		symSS = RInstall("SS", mdatass);
		symDF = RInstall("DF", nterms + 1);
		symRES = RInstall("RESIDUALS", mdatay);

		if (symRES == (Symbolhandle) 0 || symDF == (Symbolhandle) 0 ||
			symSS == (Symbolhandle) 0)
		{
			notInstalled = 1;
			break;
		}

		if (MODELTYPE & (CASEWEIGHTS | ITERWEIGHTS | IPF))
		{
			symWRES = RInstall("WTDRESIDUALS", mdatay);
			if (symWRES == (Symbolhandle) 0)
			{
				notInstalled = 1;
				break;
			}
		}

		if (MODELTYPE & MANOVA)
		{						/* modify dimensions for multivariate */
			setNDIMS(symSS,3);
			setDIM(symSS,1,nterms + 1);
			setDIM(symSS,2,ny);
			setDIM(symSS,3,ny);

			setNDIMS(symRES,2);
			setDIM(symRES,1,ndata);
			setDIM(symRES,2,ny);

			if (MODELTYPE & (CASEWEIGHTS | ITERWEIGHTS))
			{
				setNDIMS(symWRES,2);
				setDIM(symWRES,1,ndata);
				setDIM(symWRES,2,ny);
			}
		}
		
		DCOPY(&mdatass, *SS, &incr, DATAPTR(symSS), &incr);

		tmp = DATA(symRES);

		if (MODELTYPE & FASTANOVA)
		{
			j = 0;
			for (i = 0; i < ndata; i++)
			{ /* re-insert MISSING where squeezed out by funbalanova */
				if (!(MODELTYPE & MISSWEIGHTS) || (*MISSWTS)[i] > 0.)
				{
					(*tmp)[i] = (*RESIDUALS)[j++];
				}
				else
				{
					setMissing((*tmp)[i]);
				}
			}
		} /*if (MODELTYPE & FASTANOVA)*/
		else
		{
			DCOPY(&mdatay, *RESIDUALS, &incr, *tmp, &incr);
		}		

		if (MODELTYPE & (CASEWEIGHTS | ITERWEIGHTS | IPF))
		{
			double     d;
			
			tmp = DATA(symWRES);
			for (i = 0; i < mdatay; i++)
			{
				(*tmp)[i] = (*WTDRESIDUALS)[i];
				if(MODELTYPE & ROBUSTREG)
				{
					if((d = (*RESIDUALS)[i]) != 0 && !isMissing(d))
					{
						(*ALLWTS)[i] = (*WTDRESIDUALS)[i]/d;
					}
				}
			}
		}

		length = nterms + 1;
		DCOPY(&length, *DF, &incr, DATAPTR(symDF), &incr);

		if (!(MODELTYPE & FASTANOVA))
		{
			symHII = RInstall("HII", ndata);
			if (symHII == (Symbolhandle) 0)
			{
				notInstalled = 1;
				break;
			}
			DCOPY(&ndata, *HII, &incr, DATAPTR(symHII), &incr);
		}

		if (MODELTYPE & OLSREG)
		{
			symXTX = RInstall("XTXINV", mdataxx);
			symCOEF = RInstall("COEF", nregx);
			if (symXTX == (Symbolhandle) 0 || symCOEF == (Symbolhandle) 0)
			{
				notInstalled = 1;
				break;
			}
			setNDIMS(symXTX,2);
			setDIM(symXTX,1,nregx);
			setDIM(symXTX,2,nregx);
			DCOPY(&mdataxx, *XTXINV, &incr, DATAPTR(symXTX), &incr);
			DCOPY(&nregx, *REGCOEF, &incr, DATAPTR(symCOEF), &incr);
		} /*if (MODELTYPE & OLSREG)*/
		PREVMODELTYPE = MODELTYPE;
		break;
	} /*while(!usePrevious)   (fake while loop)*/
	
	if(printResults)
	{		
		if (MODELTYPE & OLSREG)
		{
			printregress();
			if(FactorFound && INTERRUPT == INTNOTSET)
			{
				sprintf(OUTSTR,
						"WARNING: factor%s in model treated as variate%s by %s",
						(FactorFound > 1) ? "s" : "",
						(FactorFound > 1) ? "s" : "",
						FUNCNAME);
						putOUTSTR();
			}
		} /*if (MODELTYPE & OLSREG)*/
		else
		{
			if (!(MODELTYPE & (IPF | BALANOVA)) || usePrevious)
			{
				myerrorout("WARNING: summaries are sequential");
			}
			printglm();
			if(usePrevious && FactorFound && INTERRUPT == INTNOTSET)
			{
				sprintf(OUTSTR,
						"WARNING: factor%s in model treated as variate%s by %s",
						(FactorFound > 1) ? "s" : "",
						(FactorFound > 1) ? "s" : "",
						FUNCNAME);
				putOUTSTR();
			}			
		} /*if (MODELTYPE & OLSREG) {...}else{...} */
	} /*if(printResults)*/
	*OUTSTR = '\0';
	
	if(!printResults && !silent)
	{
		myerrorout("NOTE: results may be retrieved from variables SS, DF, and RESIDUALS");
		if(MODELTYPE & DOCOEFS)
		{
			myerrorout("      Use coefs(), secoefs(), or modelinfo() to retrieve other results");
		}
	}

	if(notInstalled && !silent)
	{
		myerrorout("WARNING: not all the side effect variables could be created");
	}
	
	UNLOADSEG(getmodel);
	UNLOADSEG(screen);
	UNLOADSEG(ipf);
	UNLOADSEG(funbalanova);
	UNLOADSEG(iterglm);
	UNLOADSEG(unbalanova);
	UNLOADSEG(printregress);
	UNLOADSEG(printglm);
	UNLOADSEG(gramschmidt);

	
	RETURN (NULLSYMBOL);

  errorExit:
	if(doCleanup)
	{
		glmcleanup(0);
	}
	
	MODELTYPE = NOMODEL;

	putOUTSTR();
	emptyTrash();

	UNLOADSEG(getmodel);
	UNLOADSEG(screen);
	UNLOADSEG(ipf);
	UNLOADSEG(funbalanova);
	UNLOADSEG(iterglm);
	UNLOADSEG(unbalanova);
	UNLOADSEG(printregress);
	UNLOADSEG(printglm);
	UNLOADSEG(gramschmidt);
	
	RETURN (0);
	
}


/* NOTE: neither glmcleanup nor clearGlobals should use OUTSTR */

void glmcleanup(removeStrmodel)
int         removeStrmodel;
{
	WHERE("glmcleanup");
	
	/* get rid of misc anova globals after error */

	clearGlobals();	

	Remove("RESIDUALS");
	Remove("WTDRESIDUALS");
	Remove("HII");
	Remove("SS");
	Remove("DF");
	if(removeStrmodel)
	{
		Remove("STRMODEL");
	}
	Remove("DEPVNAME");
	Remove("TERMNAMES");
	Remove("XTXINV");
	Remove("COEF");
}

void clearGlobals()
{
	/* delete all glm related globals */
	long            i;
	long            nvars = (long) NVARS;
	WHERE("clearGlobals");
	
	PREVMODELTYPE = NOMODEL;
	USEGLMOFFSET = 0;
	NVARS = MODELINFO->nvars = -1;
	NFACTORS = MODELINFO->nfactors = -1;
	NTERMS = MODELINFO->nterms = 0;
	for(i=0;i<=MAXERRTRMS;i++)
	{
		zeroTerm(MODELINFO->errorterms[i]);
	}
/*
  Y was set to DATA(MODELVARS[0] and X[i] to DATA(MODELVARS[i+1] in
  getmodel().  However, they can be extended in funbalan.c and
  in that case should be deleted.
 */
	
	for (i = 0; i <= nvars; i++)
	{
		if(i == 0 && MODELVARS[0] != (Symbolhandle) 0)
		{
			if(Y != DATA(MODELVARS[0]))
			{
				mydisphandle((char **) Y);
			}
			Y = (double **) 0;
		}
		else if(MODELVARS[i] != (Symbolhandle) 0)
		{
			if(X[i-1] != DATA(MODELVARS[i]))
			{
				mydisphandle((char **) X[i-1]);
			}
			X[i-1] = (double **) 0;
		}
		
		Delete(MODELVARS[i]);
		MODELVARS[i] = (Symbolhandle) 0;
	} /*for (i = 0; i <= nvars; i++)*/
	
	/* Note: As of 930601, INTMODEL no longer exists */
	mydisphandle5((char **) MODEL, (char **) RESIDUALS,
				  (char **) WTDRESIDUALS, (char **) 0, (char **) 0);
	MODEL = (modelHandle) 0;
	RESIDUALS = WTDRESIDUALS = (double **) 0;

	mydisphandle5((char **) HII, (char **) SS, (char **) DF,
				  STRMODEL, TERMNAMES);
	HII = SS = DF = (double **) 0;
	TERMNAMES = STRMODEL = (char **) 0;

	mydisphandle(DEPVNAME);
	DEPVNAME = (char **) 0;
	
	mydisphandle5((char **) REGX, (char **) REGX2, (char **) XTXINV,
				  (char **) REGCOEF, (char **) REGSS);
	REGX = REGX2 = XTXINV = REGCOEF = REGSS = (double **) 0;

	mydisphandle5((char **) MISSWTS, (char **) CASEWTS, (char **) ITERWTS,
				  (char **) ALLWTS, (char **) GLMOFFSET);
	MISSWTS = CASEWTS = ITERWTS = ALLWTS = GLMOFFSET = (double **) 0;

	mydisphandle((char **) LOGITN);
	LOGITN = (double **) 0;
}


