/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#include "globals.h"

#ifdef MACINTOSH
#include "macIface.h"
#endif /*MACINTOSH*/

/*
  interrupt handlers; these were in main.c

  The scheme that was tried, and remains of which are still in place, was for
  each function to call setjmp() using macro SETUPINT(label), where label
  was a label to jump to in case a longjmp() was executed; the various
  jmp_buf's that were required were in a stack.  Since termination of
  output when the screen is full by typing 'q' is handled (except on the Mac)
  by calling longjmp(), this scheme also stopped memory leaps when output
  was terminated.  Currently fmyeol() sets INTERRUPT to PRINTABORT and
  calls intRoutine() which executes the longjmp() call.

  Here are comments from handlers.h

	If SETJMP is defined, RestartBuf[] is a stack of jmp_buf's
	RestartBuf[0] is set by setjmp() near start
	RestartBuf[j] is set at "depth" j to allow catching interrupts or
	output aborts so as to be able to clean up.
	If SETJMP is not defined, RestartBuf[0] is a single jmp_buf used in
	main().

  The original purpose of the sehem was to provide a local means to clean up
  allocated storage to prevent memory "leaks".  Unfortunately, although
  it worked most of the time, occasionally it would crash with a message
  stating the jmpbuffer was bad.  Eventually, it was disabled and the
  prevention of memory leaks handled by "registering" temporary storage
  as components of symbols of type GARBAGE in the symbol table.  These
  are automatically disposed of at cleanup time after an interrupt.

  Macintosh
  An interrupt is processed in doFile() in macMain.c where INTERRUPT is set
  to INTSET.  This can happen only in the event loop or if
  SkelChkOneEvent() is called which is what interrupted() does on the mac.
  Thus interrupt is effective only if a function explicitly calls
  interrupted(); it does not happen automatically.

  Unix & DOS
  An interrupt causes a transfer to intRoutine() which closes batch files,
  prints a message, sets INTERRUPT to INTSET, and calls longjmp().
  Under the scheme, this would have transferred back to the function which
  was active when the interrupt came.  Function interrupted() sets
  INTERRUPT to INTNOTSET (0) and returns INTNOTSET.
 
  Global variable INTERRUPT is set to FPESET on a floating point interrupt,
  to INTSET on any other interrupt.  The default value for interrupt is
  INTNOTSET with value zero.  At the "local" level, if INTERRUPT == INTNOTSET
  nothing needs to be done.  If INTERRUPT != INTNOTSET, then storage allocated
  needs to be released and an error exit taken.

*/

#ifndef MACINTOSH

/* take care of floating point errors for Unix & DOS */
#ifdef SIGNALARG
void fpeRoutine(int sig)
#else /*SIGNALARG*/
void fpeRoutine(void)
#endif /*SIGNALARG*/
{
	long      screenheight = SCREENHEIGHT;
	WHERE("fpeRoutine");
	
	SCREENHEIGHT = 0; /* make sure output will not trigger new interrupt */

	myerrorout("ERROR: floating point exception");
	closeBatch(1); /* shut down all batch files */
	SCREENHEIGHT = screenheight;
	INTERRUPT = FPESET;
	
#ifdef SIGNALARG
	intRoutine(0);
#else /*SIGNALARG*/
	intRoutine();
#endif /*SIGNALARG*/
	/* no return */
} /*fpeRoutine()*/

/* take care of SIGINT for Unix & DOS */
#ifdef SIGNALARG
void intRoutine(int sig)
#else /*SIGNALARG*/
void intRoutine(void)
#endif /*SIGNALARG*/
{
	long     screenheight = SCREENHEIGHT;
	WHERE("intRoutine");
	
	SCREENHEIGHT = 0; /* make sure any output will not trigger new interrupt */

#undef UNDEFINED__
#ifdef UNDEFINED__
	if(GUBED & 65536 && INTWHERE[RDEPTH] != (char *) 0)
	{
		PRINT("INTWHERE[%ld] = %s\n", RDEPTH, INTWHERE[RDEPTH],0,0);
	}
#endif /*UNDEFINED__*/

	if(INTERRUPT == INTNOTSET)
	{ /* actual interrupt */
		SCREENHEIGHT = 0; /* make sure output will not trigger new interrupt */
		myeol();
		myeol();
		myerrorout("  *** INTERRUPT ***");
		myeol();
		SCREENHEIGHT = screenheight;
		INTERRUPT = INTSET;
		closeBatch(1); /* shutdown all batch files */
	}
	SCREENHEIGHT = screenheight;
	NLINES = 0;
	/* 
	  It is the responsibility of each function that sets up RestartBuf
	  to increment and decrement RDEPTH.  This should be automatic if
	  macros SETUPINT and RETURN are used.
	*/
	longjmp(RestartBuf[RDEPTH], 0);
}

#ifdef DJGPP
#include <go32.h>
#endif /* DJGPP */

long interrupted(void)
{
#ifdef DJGPP        
	WHERE("interrupted");

	if( _go32_was_ctrl_break_hit() )
	{
		INTERRUPT = INTNOTSET;
		intRoutine();
	}
#endif /* DJGPP */
	return (INTERRUPT);
}
#else /*MACINTOSH*/

/*
	Temporary idle procedure that will be executed at most once
*/

static pascal void
CheckOneEvent(void)
{
	SkelStopEventLoop();
}

/*
	Routine to check one event and route it, without looping through
	window handlers.
*/

pascal void
SkelChkOneMaskedEvent (Integer mask, LongInt waitTime)
{
	Integer         oldTSMask;
	LongInt         oldFgWaitTime, oldBgWaitTime;
	SkelIdleProcPtr oldIdle;

	oldTSMask = SkelGetEventMask(); /* and event mask */
	SkelGetWaitTimes(&oldFgWaitTime, &oldBgWaitTime);
	oldIdle = SkelGetIdle();/* get current idle proc */
	
	/* install new idle & mask */
	SkelSetEventMask(mask);
	SkelSetWaitTimes(waitTime, oldBgWaitTime);
	SkelSetIdle(CheckOneEvent);

	SkelEventLoop();        /* handle given event types only */

	/* restore stuff that was changed */
	SkelSetEventMask(oldTSMask);
	SkelSetWaitTimes(oldFgWaitTime, oldBgWaitTime);
	SkelSetIdle(oldIdle);
} /*SkelChkOneMaskedEvent()*/

/*
	Interface to check for one event of any type using default wait time
	Used primarily to check for interrupts, but also allows user access to
	desk accessories, multi-finder, etc.
*/

pascal void
SkelChkOneEvent (void)
{
	LongInt       fgWaitTime;

	SkelGetWaitTimes(&fgWaitTime, (LongInt) nil);
	SkelChkOneMaskedEvent (everyEvent, fgWaitTime);

} /*SkelChkOneEvent()*/

/*
	If Command I or Command period or Interrupt on File menu,
	set INTERRUPT to INTSET (defined in handlers.h as 0x10000)
	Test of whether an interrupt has occurred should be
	  if(interrupted() != INTNOTSET){...process interrupt...}
	
	The only place where INTERRUPT should be set to INTNOTSET is in
	cleanitup() in mainpars.c
*/

long interrupted(void)
{
	WHERE("interrupted");

	if(INTERRUPT == INTNOTSET)
	{
		SkelChkOneMaskedEvent (everyEvent, InterruptSleep);
	}

	return (INTERRUPT);
}
#endif /*MACINTOSH*/

