/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Iterfns
#else
#define __SEG__ Iterfns
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <math.h>

#define newresidH     RESIDUALS       
#define fitH          WTDRESIDUALS    
#define lastresidH    HII             

/*
  This module contains three routines for each of the iterative least
  squares algorithms.   fit[i] is (*fitH)[i], newresid[i] is (*newresidH)[i],
  and lastresid[i] is (*lastresidH)[i].

  In the following descriptions xxxxx stands for robust, poisson, or logit

  xxxxxvar() updates ALLWTS, taking into account MISSWTS, CASEWTS, and
  the current fit.

  xxxxxlink(INITIAL) initializes fit[], newresid[], and lastresid[]
  xxxxxlink(STEP) updates fit[], newresid[], and lastresid[]
	Except for robustlink(STEP), xxxxxlink(op) always sets newresid[i]
	and lastresid[i] the same.

  xxxxxdev(INITIAL) computes initial value of deviance or other criteria
  xxxxxdev(STEP) computes new value of deviance or other criteria

  xxxxxresid() puts the actual residual (*Y)[i] - fit[i] in newresid[i]
  and puts the previous value of newresid[i] in fit[i]
*/

extern void     itergs();

double          CURRENTRSC;
double          ROBUSTVARK; /* this is K from page 40 of Huber 1977 */
double          REGDIM;

void            robustvar(op)
long            op;
{
	long            i;
	long            ndata = (long) NDATA;
	double         *allwts = *ALLWTS;
	double          weight;
	
	for (i = 0; i < ndata; i++)
	{
		weight = 1.0;
		if (MODELTYPE & MISSWEIGHTS)
		{
			weight *= (*MISSWTS)[i];
		}
		if (MODELTYPE & CASEWEIGHTS)
		{
			weight *= sqrt((*CASEWTS)[i]);
		}
		/* leave wts 1 , modify residuals */
		allwts[i] = weight;
	} /*for (i = 0; i < ndata; i++)*/
} /*robustvar()*/

void            robustlink(op)
long            op;
{
	long            i, j, k;
	double          z, pred;
	double          sumPsiSq;
	double          beta = .35, c = .75;
	double         *newresid = *newresidH;
	double         *lastresid = *lastresidH;
	double         *fit = *fitH;
	double         *allwts = *ALLWTS, *y = *Y;
	long            ndata = (long) NDATA;

	if(op == INITIAL)
	{
		/* initialize CURRENTRSC */
		k = 0;
		for (i = 0; i < ndata; i++)
		{
			if (!(MODELTYPE & MISSWEIGHTS) || (*MISSWTS)[i] != 0.0)
			{
				newresid[k++] = y[i];
			}
		} /*for (i = 0; i < ndata; i++)*/

		sortquick(newresid, k);

		i = 3 * k / 4 - 1;
		j = k / 4 - 1;
		CURRENTRSC = (newresid[i] - newresid[j]) / 1.345;
		if (CURRENTRSC == 0.0)
		{
			CURRENTRSC = 1e-10;
		}
#ifdef UNDEFINED__
		if(modeltermEqual(modelterm(MODEL,0),(modelPointer) NULLTERM))
		{ /* constant term in model, set pred to median */
			i = (k - 1)/ 2;
			j = k/2;
			pred = .5*(newresid[i] + newresid[j]);
		}
		else
#endif /*UNDEFINED__*/
		{
			pred = 0.0;
		}
		
		for (j = 0; j < 5; j++)
		{
			sumPsiSq = 0.;
			for (i = 0; i < ndata; i++)
			{
				if (!(MODELTYPE & MISSWEIGHTS) || (*MISSWTS)[i] != 0.0)
				{
					z = fabs(y[i] - pred) / CURRENTRSC;
					if (z >= c)
					{
						z = c;
					}
					sumPsiSq += z * z;
				}
			} /*for (i = 0; i < ndata; i++)*/
			CURRENTRSC = sqrt(sumPsiSq / (NOTMISSING * beta)) * CURRENTRSC;
			if (CURRENTRSC == 0.0)
			{
				CURRENTRSC = 1e-10;
			}
		} /*for (j = 0; j < 5; j++)*/

		for (i = 0; i < ndata; i++)
		{
			z = (y[i] - pred) / CURRENTRSC;
			if (z >= c)
			{
				z = c;
			}
			else if (z <= -c)
			{
				z = -c;
			}
			z *= CURRENTRSC;
			fit[i] = 0.0; /* old predicted values*/
			newresid[i] = pred + z; /* modified eta */
			lastresid[i] = z;
		} /*for (i = 0; i < ndata; i++)*/
	} /*if(op == INITIAL)*/
	else
	{ /*op == STEP*/
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] != 0.0)
			{
				fit[i] += lastresid[i] - newresid[i];
				pred = fit[i];
				/* old eta hat */
				z = (y[i] - pred) / CURRENTRSC;
				if (z >= c)
				{
					z = c;
				}
				else if (z <= -c)
				{
					z = -c;
				}
				z *= CURRENTRSC;
				newresid[i] = pred + z; /* modified eta */
				lastresid[i] = z;
			} /*if (allwts[i] != 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/

		/* update CURRENTRSC */
		sumPsiSq = 0.;
		ROBUSTVARK = 0.;
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] > 0.0)
			{
				pred = fit[i];
				z = fabs(y[i] - pred) / CURRENTRSC;
				if (z >= c)
				{
					z = c;
				}
				else
				{
					ROBUSTVARK++;
				}
				sumPsiSq += z * z;
			} /*if (allwts[i] > 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
		ROBUSTVARK /= NOTMISSING;
		ROBUSTVARK = (1.0+(REGDIM/NOTMISSING)*(1.0-ROBUSTVARK)/ROBUSTVARK)/
			ROBUSTVARK;
		/* this is K from page 40 of Huber 1977 */
		if (NOTMISSING == REGDIM)
		{
			CURRENTRSC = 1e-10;
		}
		else
		{
			CURRENTRSC = sqrt(sumPsiSq/
							  ((NOTMISSING - REGDIM) * beta)) * CURRENTRSC;
		}
		if (CURRENTRSC == 0.0)
		{
			CURRENTRSC = 1e-10;
		}
	} /*if(op == INITIAL){}else{}*/

} /*robustlink()*/


/*
  Uses rho-function for eq. 7.14 of Huber (1981)
*/
double          robustdev(op)
long            op;
{
	long            i;
	double          sumrho, z;
	double         *allwts = *ALLWTS, *y = *Y;
	double         *fit = *fitH;
	double          beta = .35, c = .75;
	double          addcon1 = .5*beta, addcon2 = -.5*c*c + addcon1;
	long            ndata = (long) NDATA;

	sumrho = 0.;

	for (i = 0; i < ndata; i++)
	{
		if (allwts[i] != 0.0)
		{
			z = fabs(y[i] - fit[i]) / CURRENTRSC;
			if (z < c)
			{
				sumrho += 0.5 * z * z + addcon1;
			}
			else
			{
				sumrho += c * z + addcon2;
			}
		} /*if (allwts[i] != 0.0)*/
	} /*for (i = 0; i < ndata; i++)*/
	return (sumrho * CURRENTRSC);
} /*robustdev()*/

void            robustresid()
{
	long            i;
	double         *newresid = *newresidH;
	double         *fit = *fitH;
	double         *allwts = *ALLWTS, *y = *Y;
	double        **tmpH;
	long            ndata = (long) NDATA;

	for (i = 0; i < ndata; i++)
	{
		if (allwts[i] == 0)
		{
			setMissing(newresid[i]);
		}
		else
		{
			/* modify predicted to residual in place */
			fit[i] = y[i] - fit[i];
		}
	} /*for (i = 0; i < ndata; i++)*/
	tmpH = fitH;
	fitH = newresidH;
	newresidH = tmpH;
} /*robustresid()*/

void            logitvar(op)
long            op;
{
	long            i;
	long            ndata = (long) NDATA;
	double         *allwts = *ALLWTS, *logitn = *LOGITN;
	double         *fit = *fitH;
	double          pred;
	
	for (i = 0; i < ndata; i++)
	{
		allwts[i] = 1.0;
		if (MODELTYPE & MISSWEIGHTS)
		{
			allwts[i] *= (*MISSWTS)[i];
		}
		if (MODELTYPE & CASEWEIGHTS)
		{
			allwts[i] *= sqrt((*CASEWTS)[i]);
		}
		pred = fit[i];
		if (allwts[i] > 0.0 && pred > 0.0)
		{
			allwts[i] *= sqrt(logitn[i] * (pred * (1. - pred)));
		}
	} /*for (i = 0; i < ndata; i++)*/
} /*logitvar()*/

void            logitresid()
{
	long            i;
	double         *newresid = *newresidH;
	double         *fit = *fitH;
	double         *allwts = *ALLWTS, *y = *Y, *logitn = *LOGITN;
	double        **tmpH;
	long            ndata = (long) NDATA;

	for (i = 0; i < ndata; i++)
	{
		if (allwts[i] == 0)
		{
			setMissing(newresid[i]);
		}
		else
		{
			/* modify predicted to residual in place */
			fit[i] = (y[i] - logitn[i] * fit[i]);
		}
	} /*for (i = 0; i < ndata; i++)*/
	/* swap fitH (WTDRESIDUALS) and rewresidH (RESIDUALS) */
	tmpH = fitH;
	fitH = newresidH;
	newresidH = tmpH;
} /*logitresid()*/

/*
   Note: the only difference between when op == INITIAL and op != INITIAL
   has to do with the values used for pred
*/

double          logitdev(op)
long            op;
{
	long            i;
	double         *fit = *fitH;
	double          tmp, logitni, p, pred;
	double         *y = *Y, *logitn = *LOGITN;
	double         *allwts = *ALLWTS;
	long            ndata = (long) NDATA;

	tmp = 0.;
	if (op == INITIAL)
	{
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] != 0.0)
			{
				pred = (USEGLMOFFSET) ? exp((*GLMOFFSET)[i]) : 1.0;
				pred /= 1.0+pred;
				logitni = logitn[i];
				p = y[i]/logitni;
				if(p == 0.0) 
				{
					tmp += logitni*log(1.0/(1.0-pred));
				}
				else if(p == 1.0) 
				{
					tmp += logitni * log(1.0/pred);
				} 
				else 
				{
					tmp += logitni*
						(p*log(p/pred) + (1.0 -p)*log((1.0 - p)/(1.0 - pred)));
				}
			} /*if (allwts[i] != 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
	} /* if (op == INITIAL) */
	else
	{
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] != 0.0)
			{
				logitni = logitn[i];
				p = y[i]/logitni;
				pred = fit[i];
				if (p == 0.0)
				{
					tmp += logitni * log(1.0 / (1.0 - pred));
				}
				else if (p == 1.0)
				{
					tmp += logitni * log(1.0 / pred);
				}
				else
				{
					tmp += logitni *
						(p*log(p/pred) + (1.0 - p)*log((1.0 - p)/(1. - pred)));
				}
			} /*if (allwts[i] != 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
	} /* if (op == INITIAL){}else{} */
	return (2.0 * tmp);
} /*logitdev()*/

void            logitlink(op)
long            op;
{
	long            i;
	double         *newresid = *newresidH;
	double         *lastresid = *lastresidH;
	double         *fit = *fitH;
	double         *allwts = *ALLWTS, *y = *Y, *logitn = *LOGITN;
	double          pred, resid;
	long            ndata = (long) NDATA;

	if(op == INITIAL)
	{
		for (i = 0; i < ndata; i++)
		{
			/* predicted values */
			fit[i] = 0.5;
			resid = 4. * (y[i] / logitn[i] - 0.5);
			if(USEGLMOFFSET)
			{
				resid -= (*GLMOFFSET)[i];
			}
			lastresid[i] = newresid[i] = resid;
		} /*for (i = 0; i < ndata; i++)*/
	} /*if(op == INITIAL)*/
	else
	{ /*op == STEP */
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] != 0.0)
			{
				/* new eta hat */
				pred = lastresid[i] - newresid[i] / allwts[i];
				if(USEGLMOFFSET)
				{
					pred += (*GLMOFFSET)[i];
				}
				/* clamp to prevent going off to infinity */
				if (pred > 10.)
				{
					pred = 10.;
				}
				else if (pred < -10.)
				{
					pred = -10.;
				}
				fit[i] = pred = 1.0/ (1. + exp(-pred));
				/* modified eta */

				resid = log(pred / (1. - pred)) +
					(y[i] / logitn[i] - pred) / (pred * (1. - pred));
				if(USEGLMOFFSET)
				{
					resid -= (*GLMOFFSET)[i];
				}
				lastresid[i] = newresid[i] = resid;
			} /*if (allwts[i] != 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
	} /*if(op == INITIAL){}else{}*/
} /*logitlink()*/

void            poissonvar(op)
long            op;
{
	long            i;
	long            ndata = (long) NDATA;
	double         *allwts = *ALLWTS;
	double         *fit = *fitH;
	double          pred;
	
	for (i = 0; i < ndata; i++)
	{
		allwts[i] = 1.0;
		if (MODELTYPE & MISSWEIGHTS)
		{
			allwts[i] *= (*MISSWTS)[i];
		}
		if (MODELTYPE & CASEWEIGHTS)
		{
			allwts[i] *= sqrt((*CASEWTS)[i]);
		}
		pred = fit[i];
		if (allwts[i] > 0.0 && pred > 0.0)
		{
			allwts[i] *= sqrt(pred);
		}
	} /*for (i = 0; i < ndata; i++)*/
} /*poissonvar()*/

double          poissondev(op)
long            op;
{
	long            i;
	double         *fit = *fitH;
	double          tmp,pred, yi;
	double         *allwts = *ALLWTS, *y = *Y;
	long            ndata = (long) NDATA;

	tmp = 0.;
	if (op == INITIAL)
	{
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] != 0.0)
			{
				pred = (USEGLMOFFSET) ? exp((*GLMOFFSET)[i]) : 1.;
				if((yi = y[i]) > 0.0 ) 
				{
					tmp += yi * log(yi/pred) - yi + pred;
				} 
				else 
				{
					tmp += pred;
				}
			} /*if (allwts[i] != 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
	} /*if (op == INITIAL)*/
	else
	{
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] != 0.0)
			{
				pred = fit[i];
				if ((yi = y[i]) > 0.0)
				{
					tmp += yi * log(yi / pred) - yi + pred;
				}
				else
				{
					tmp += pred;
				}
			} /*if (allwts[i] != 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/
	} /*if (op == INITIAL){}else{}*/
	
	return (2.0 * tmp);
} /*poissondev()*/

void            poissonlink(op)
long            op;
{
	long            i, j;
	long            ndata = (long) NDATA, nregx = (long) NREGX;
	long            nvals = ndata*nregx;
	double         *newresid = *newresidH;
	double         *lastresid = *lastresidH;
	double         *fit = *fitH;
	double         *regx, *regx2;
	double         *y = *Y, *xij;
	double         *allwts = *ALLWTS;
	double          tmp, pred, resid;
	
	if (op == INITIAL)
	{
		regx = *REGX;
		regx2 = *REGX2;
		for (i = 0; i < nvals; i++)
		{
			*regx++ = *regx2++;
		} /*for (i = 0; i < nvals; i++)*/

		regx = *REGX;
		for (i = 0; i < ndata; i++)
		{
			newresid[i] = y[i];
			if (MODELTYPE & MISSWEIGHTS && (*MISSWTS)[i] == 0.0)
			{
				newresid[i] = 0.0;
				xij = regx + i;
				for (j = 0; j < nregx; j++)
				{
					*xij = 0.;
					xij += ndata;
				}
			}
		} /*for (i = 0; i < ndata; i++)*/

		gramschmidt(DONTNOTIFY | DOSWEEP | DONTDOSS | DONTDOCOEFS |
					DONTDORINV);
		if(INTERRUPT != INTNOTSET)
		{
			goto interruptExit;
		}
		for (i = 0; i < ndata; i++)
		{
			if( (MODELTYPE & MISSWEIGHTS) && (*MISSWTS)[i] == 0.)
			{
				fit[i] = 1.;
				newresid[i] = 0.0;
			} /*if( (MODELTYPE & MISSWEIGHTS) && (*MISSWTS)[i] == 0.0)*/
			else
			{
				/* predicted values */
				pred = y[i] - newresid[i];
				if (pred <= 0.5)
				{
					pred = 0.5;
					/* modified eta */
				}
				fit[i] = pred;
				resid = log(pred) + (y[i] - pred) / pred;
				if(USEGLMOFFSET)
				{
					resid -= (*GLMOFFSET)[i];
				}
				lastresid[i] = newresid[i] = resid;
			}/*if( (MODELTYPE & MISSWEIGHTS) && (*MISSWTS)[i] == 0.0){}else{}*/
		} /*for (i = 0; i < ndata; i++)*/
	} /*if (op == INITIAL)*/
	else
	{ /*op == STEP*/
		for (i = 0; i < ndata; i++)
		{
			if (allwts[i] == 0.0)
			{
				lastresid[i] = newresid[i] = fit[i] = 0.0;
			}
			else
			{
				/* new eta hat */
				tmp = lastresid[i] - newresid[i] / allwts[i];
				if(USEGLMOFFSET)
				{
					tmp += (*GLMOFFSET)[i];
				}
				fit[i] = pred = exp(tmp);
				/* modified eta */
				resid = log(pred) + (y[i] - pred) / pred;
				if(USEGLMOFFSET)
				{
					resid -= (*GLMOFFSET)[i];
				}
				lastresid[i] = newresid[i] = resid;
			} /*if (allwts[i] == 0.0)*/
		} /*for (i = 0; i < ndata; i++)*/		
	} /*if (op == INITIAL){}else{}*/
	/* fall through*/
  interruptExit:
  	;
}  /*poissonlink()*/

void            poissonresid()
{
	long            i;
	double         *fit = *fitH;
	double         *newresid = *newresidH;
	double        **tmpH;
	double         *allwts = *ALLWTS, *y = *Y;
	long            ndata = (long) NDATA;

	for (i = 0; i < ndata; i++)
	{
		if (allwts[i] == 0)
		{
			setMissing(newresid[i]);
		}
		else
		{
			/* modify predicted to residual in place */
			fit[i] = y[i] - fit[i];
		}
	}
	tmpH = fitH;
	fitH = newresidH;
	newresidH = tmpH;
} /*poissonresid()*/
