/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

/*
  Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu)
*/

/* does not require stdio.h; requires math.h (included by linpack.h)
   to declare sqrt() and fabs()
*/

#include "linpack.h"  /* prototypes and macros for names */
#include "blas.h"  /* prototypes and macros for names */

#ifndef INT
#define INT      long
#endif /*INT*/

/*
  dqrdc uses Householder transformations to compute the qr
  factorization of an n by p matrix x.  Column pivoting
  based on the 2-norms of the reduced columns may be
  performed at the users option.

  On entry

     x       double precision(ldx,p), where ldx .ge. n.
             x contains the matrix whose decomposition is to be
             computed.

     ldx     integer.
             ldx is the leading dimension of the array x.

     n       integer.
             n is the number of rows of the matrix x.

     p       integer.
             p is the number of columns of the matrix x.

     jpvt    integer(p).
             jpvt contains integers that control the selection
             of the pivot columns.  The k-th column x(k) of x
             is placed in one of three classes according to the
             value of jpvt(k).

                If jpvt(k) .gt. 0, then x(k) is an initial
                                   column.

                If jpvt(k) .eq. 0, then x(k) is a free column.

                If jpvt(k) .lt. 0, then x(k) is a final column.

             Before the decomposition is computed, initial columns
             are moved to the beginning of the array x and final
             columns to the end.  Both initial and final columns
             are frozen in place during the computation and only
             free columns are moved.  At the k-th stage of the
             reduction, if x(k) is occupied by a free column
             it is interchanged with the free column of largest
             reduced norm.  jpvt is not referenced if
             job .eq. 0.

     work    double precision(p).
             work is a work array.  work is not referenced if
             job .eq. 0.

     job     integer.
             job is an integer that initiates column pivoting.
             If job .eq. 0, no pivoting is done.
             If job .ne. 0, pivoting is done.

  On return

     x       x contains in its upper triangle the upper
             triangular matrix r of the qr factorization.
             Below its diagonal x contains information from
             which the orthogonal part of the decomposition
             can be recovered.  Note that if pivoting has
             been requested, the decomposition is not that
             of the original matrix x but that of x
             with its columns permuted as described by jpvt.

     qraux   double precision(p).
             qraux contains further information required to recover
             the orthogonal part of the decomposition.

     jpvt    jpvt(k) contains the index of the column of the
             original matrix that has been interchanged into
             the k-th column, if pivoting was requested.

  Linpack. this version dated 08/14/78 .
  G.W. Stewart, University of Maryland, Argonne National Lab.

*/

/*
  In this C version, with indices starting at zero, jpvt[k] contains the
  index+1 of the interchanged column, i.e., just what the Fortran version
  uses.
*/


/*
  dqrdc uses the following functions and subprograms.

  blas daxpy,ddot,dscal,dswap,dnrm2
  fortran dabs,dmax1,min0,dsqrt
*/

/*
  Whether using Fortran or C storage conventions, it is assumed that
  X(i,j) is the element in row i and column j of array x.

  If C storage conventions are used, ldx is the size of the trailing 
  dimension of array x
*/

#ifdef FORTRAN  /* 1st subscript changes fastest */
#define INCI 1
#define INCJ ldx
#define PX(I,J) (x + (I) + (J)*INCJ) /* &X(I,J) */
#else           /* 2nd subscript changes fastest */
#define INCI ldx
#define INCJ 1
#define PX(I,J) (x + (I)*INCI + (J)) /* &X(I,J) */
#endif
#define X(I,J) (*PX(I,J))

void DQRDC(x,pldx,pn,pp,qraux,jpvt,work,pjob)
INT           *pldx, *pn, *pp, *pjob;
INT            jpvt[];
double         x[],qraux[],work[];
/* double precision x(ldx,1), qraux(1), work(1)*/
{
	/*      internal variables */

	register INT ldx = *pldx;
	INT          n = *pn, p = *pp, job = *pjob;
	INT          j,jp,l,lup,maxj,pl,pu;
	double       maxnrm,tt;
	double       nrmxl, t;
	double       temp;
	INT          negj,swapj;
	INT          length, inci = INCI;

	pl = 0;
	pu = -1;
	if (job != 0)
	{
		/*
		  Pivoting has been requested.  Rearrange the columns
		  according to jpvt.
		*/

		for(j=0;j<p ;j++)
		{
			swapj = (jpvt[j] > 0);
			negj = (jpvt[j] < 0);
			jpvt[j] = (negj) ? -(j+1) : j+1 ;/* Fortran style index returned */
			if (swapj)
			{
				if (j != pl)
				{
					DSWAP(&n,PX(0,pl),&inci,PX(0,j),&inci);
				}
				jpvt[j] = jpvt[pl];
				jpvt[pl] = j+1;
				pl++;
			}
		} /* for j */
		pu = p-1;
		for(j=pu;j >= 0;j--)
		{
			if (jpvt[j]<0)
			{
				jpvt[j] = -jpvt[j];
				if (j != pu)
				{
					DSWAP(&n,PX(0,pu),&inci,PX(0,j),&inci);
					jp = jpvt[pu];
					jpvt[pu] = jpvt[j];
					jpvt[j] = jp;
				}
				pu--;
			}
		} /* for j */
	} /* if(job != 0) */

	/* compute the norms of the free columns. */

	for(j=pl;j<=pu ;j++)
	{
		work[j] = qraux[j] = DNRM2(&n,PX(0,j),&inci);
	}

	/* perform the Householder reduction of x. */

	lup = (n < p) ? n : p;
	for(l=0;l<lup ;l++)
	{
		if (l>=pl && l<pu)
		{
			/*
			  locate the column of largest norm and bring it
			  into the pivot position.
			*/

			maxnrm = 0.0;
			maxj = l;
			for(j=l;j<=pu ;j++)
			{
				if (qraux[j] > maxnrm)
				{
					maxnrm = qraux[j];
					maxj = j;
				}
			}

			if (maxj != l)
			{
				DSWAP(&n,PX(0,l),&inci,PX(0,maxj),&inci);
				qraux[maxj] = qraux[l];
				work[maxj] = work[l];
				jp = jpvt[maxj];
				jpvt[maxj] = jpvt[l];
				jpvt[l] = jp;
			}
		}
		qraux[l] = 0.0;

		if (l < n-1)
		{ /* Compute the Householder transformation for column l. */
			length = n-l;
			nrmxl = DNRM2(&length,PX(l,l),&inci);
			if (nrmxl != 0.0)
			{
				if( X(l,l) < 0)
				{
					nrmxl = -nrmxl;
				}
				
				temp = 1.0/nrmxl;
				DSCAL(&length,&temp,PX(l,l),&inci);
				X(l,l)++;

				/*
				  apply the transformation to the remaining columns,
				  updating the norms.
				*/

				for(j=l+1;j<p ;j++)
				{
					length = n-l;
					t = -DDOT(&length,PX(l,l),&inci,PX(l,j),&inci)/X(l,l);
					DAXPY(&length,&t,PX(l,l),&inci,PX(l,j),&inci);
					if (j>=pl && j<=pu && qraux[j] != 0.0)
					{
						temp = fabs(X(l,j))/qraux[j];
						tt = 1.0-temp*temp;
						tt = (tt < 0.0) ? 0.0 : tt;
						t = tt;
						temp = qraux[j]/work[j];
						tt = 1.0+0.05*tt*temp*temp;
						if (tt != 1.0)
						{
							qraux[j] *= sqrt(t);
						}
						else
						{
							length--;
							qraux[j] = DNRM2(&length,PX(l+1,j),&inci);
							work[j] = qraux[j];
						}
					}
				} /* for j */

				/* save the transformation. */

				qraux[l] = X(l,l);
				X(l,l) = -nrmxl;
			} /* if (nrmxl != 0.0) */
		} /* if (l != n-1)  */
	} /* for l */
}

