/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Movavg
#else
#define __SEG__ Movavg
#endif /*MPW3*/
#endif /*MPW*/


#include "globals.h"

#define      MIN(A,B)  (((A) < (B)) ? (A) : (B))
#define      MAX(A,B)  (((A) > (B)) ? (A) : (B))

#define AUTOREG 0x00
#define REVERSE 0x01
#define MOVAVG  0x02

/*
  qauto() -- working function for AR or MA operator

  Vector coefs is assumed to contain autoregression coefficients phi or
  moving average coefficients theta
  Each column of b is assumed to contain a set of 'innovations'

  It is assumed that destination c has already been initialized
  rows of c before i1 or after i2 (if any) are not changed
  but for autoreg, rows before i1 (if iop == 1) or after i2 (if iop == 2)
  are used as starting values

  By C. Bingham, Applied Statistics, U. of Minnesota.
  version 901220 adapted from a version used in matter on the Cyber
  910401 cosmetic changes only
  Translated to C, 920306 for use in Matter2/MacAnova

  subroutine to perform 'autoreg', 'movavg'
  iop = 0x00   autoreg
  iop = 0x01   autoreg  reverse
  iop = 0x02   movavg
  iop = 0x03   movavg   reverse
*/

static void qauto(double coefs[],double b[],double c[],
				  double scr1[],double scr2[],
				  long m1, long m2, long n2, long i1, long i2, int iop)
/*
Matrices in Fortran ordering
double coefs[m1],b[m2][n2],c[m2][n2];
double scr1[m2],scr2[m2];
*/
{
	int        ma = iop & MOVAVG, reverse = iop & REVERSE;
	double     value;
	double     signer;
	double     *bj = b, *cj = c;
	long       inc,k1,kstart,kend,i,j,k,l,ii1,ii2;
	long       nmany;
	WHERE("qauto");
	
	signer = (ma) ? -1.0 : 1.0;
	ii1 = MIN(i1,i2);
	ii2 = MAX(i1,i2);
	nmany = ii2 - ii1 + 1;
	if (!reverse)
	{
		inc = 1;
		kstart = ii1;
		kend = -1;
	}
	else
	{
		inc = -1;
		kstart = ii2;
		kend = m2;
	}

	for(j=0;j<n2 ;j++)
	{ /* loop over columns */
		for(i=0;i<m2 ;i++)
		{ /* copy x to scr1 */
			scr1[i] = bj[i];
		}
		for(i=0;i<ii1 ;i++)
		{ /* copy starting values, if any, to scr2 */
			scr2[i] = cj[i];
		}
		for(i=ii2+1;i<m2 ;i++)
		{ /* copy starting values, if any, to scr2 */
			scr2[i] = cj[i];
		}
		k1 = kstart;
		for(i=0;i<nmany ;i++)
		{
			scr2[k1] = scr1[k1];
			k = k1;
			for(l=0;l<m1 ;l++)
			{
				k -= inc;
				if (k==kend)
				{
					break;
				}
				value = coefs[l] * ((ma) ? scr1[k] : scr2[k]);
				scr2[k1] += signer * value;
			} /*for(l=0;l<m1 ;l++)*/
			k1 += inc;
		} /*for(i=0;i<nmany ;i++)*/
		for(i=ii1;i<=ii2 ;i++)
		{ /* copy computed values to output */
			cj[i] = scr2[i];
		}
		bj += m2;
		cj += m2;
	} /*for(j=0;j<n2 ;j++)*/
} /*qauto()*/

/*
  operations movavg and autoreg
  functions for applying autoregressive or moving average operator
  to (segments of) columns of argument.  They may be used to generate
  autoregressive or moving average time series if the columns of the
  argument are innovation series.
  usage:  y <- autoreg(phi,x)                     iop = 0x00
          y <- autoreg(phi,x,reverse:T)           iop = 0x01 REVERSE
          y <- movavg(theta,x)                    iop = 0x02 MOVAVG
          y <- movavg(theta,x,reverse:T)          iop = 0x03 REVERSE | MOVAVG
     or:
  usage:  y <- autoreg(phi,x,limits:c(i1,i2),start:y1)           iop = 0x00
          y <- autoreg(phi,x,reverse:T,limits:c(i1,i2),start:y1) iop = 0x01
          y <- movavg(theta,x,limits:c(i1,i2),start:y1)          iop = 0x02
          y <- movavg(theta,x,reverse:T,limits:c(i1,i2),start:y1)iop = 0x03

*/

Symbolhandle movavg(list)
Symbolhandle         list;
{
	Symbolhandle     coefs, x, start = (Symbolhandle) 0;
	Symbolhandle     result = (Symbolhandle) 0;
	Symbolhandle     arg;
	double         **scratch = (double **) 0;
	long             limits[2], dims[2], nrows,ncols, ncoefs;
	long             i, j, nargs = NARGS(list);
	long             iop, reverse = 0;
	char            *keyword;
	WHERE("movavg");
	TRASH(1,errorExit);

	iop = (strcmp(FUNCNAME,"movavg") == 0) ? MOVAVG : AUTOREG;
	limits[0] = limits[1] = 0;
	
	if(nargs < 2)
	{
		badNargs(FUNCNAME,-1002);
		goto errorExit;
	}
	if(nargs > 5)
	{
		badNargs(FUNCNAME,-5);
	}
	
	for(i=0;i<2;i++)
	{
		if(!argOK(COMPVALUE(list,i),REAL,i+1))
		{
			goto errorExit;
		}
	}
	coefs = COMPVALUE(list,0);
	x = COMPVALUE(list,1);
	OUTSTR[0] = '\0';
	if(!isVector(coefs))
	{
		sprintf(OUTSTR,"ERROR: %s must be real vector for %s",
				(iop & MOVAVG) ? "theta" : "phi" , FUNCNAME);
	}
	else if(!isMatrix(x,dims))
	{
		sprintf(OUTSTR, "ERROR: 2nd argument to %s must vector or matrix",
				FUNCNAME);
	}
	if(anyMissing(coefs) || anyMissing(x))
	{
		sprintf(OUTSTR,
				"ERROR: 1st or 2nd argument to %s contains missing value(s)",
				FUNCNAME);
	}
	else
	{
		ncoefs = symbolSize(coefs);
		nrows = dims[0];
		ncols = dims[1];
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	for(i=2;i<nargs;i++)
	{
		arg = COMPVALUE(list,i);
		keyword = isKeyword(arg);
		if(keyword == (char *) 0)
		{
			goto usage;
		}
		if(strcmp(keyword,"reverse") == 0)
		{
			if(!argOK(arg,LOGIC,i+1) || !isScalar(arg))
			{
				goto usage;
			}
			if((long) DATAVALUE(arg,0))
			{
				reverse = (int) DATAVALUE(arg,0);
			}
		}
		else if(strcmp(keyword,"limits") == 0)
		{
			if(!argOK(arg,REAL,i+1) || !isVector(arg) ||
			   symbolSize(arg) != 2)
			{
				goto usage;
			}
			for(j=0;j<2;j++)
			{
				if(DATAVALUE(arg,j) != floor(DATAVALUE(arg,j)) ||
				   (limits[j] = (long) DATAVALUE(arg,j)) < 1 ||
				   limits[j] > nrows)
				{
					myerrorout("ERROR: elements of limits must be integers between 1 and nrows(x)");
					goto usage;
				}
			}
			if(limits[0] >= limits[1])
			{
				myerrorout("ERROR: limits[1] >= limits[1]");
				goto usage;
			}
		}
		else if(strcmp(keyword,"start") == 0)
		{
			if(!argOK(arg,REAL,i+1) || !isMatrix(arg,dims) ||
			   dims[0] != nrows || dims[1] != ncols)
			{
				myerrorout("ERROR: start must be the same size and shape as x");
				goto usage1;
			}
			start = arg;
			if(anyMissing(start))
			{
				sprintf(OUTSTR,"ERROR: missing value(s) in start for %s",
						FUNCNAME);
				goto errorExit;
			}
		}
		else
		{
			sprintf(OUTSTR,"ERROR: unrecognized keyword %s",keyword);
			goto usage;
		}
	}
	if(start != (Symbolhandle) 0 && limits[0] == 0)
	{
		sprintf(OUTSTR,"ERROR: must specify 'limits' when 'start' is used");
	}
	else if(start == (Symbolhandle) 0 &&
			(limits[0] > 1 || (limits[1] > 0 && limits[1] < nrows)))
	{
		sprintf(OUTSTR,"ERROR: must specify 'start' when 'limits' is used");
	}
	else if(limits[0] == 0)
	{
		limits[0] = 1;
		limits[1] = nrows;
	}
	
	if(*OUTSTR)
	{
		putOUTSTR();
		goto usage1;
	}
	
	if(!getScratch(scratch,0,2*nrows,double))
	{
		goto errorExit;
	}

	if(start != (Symbolhandle) 0)
	{
		result = RInstall(SCRATCH,0);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		if(Copy(start,result) == 0)
		{
			goto errorExit;
		}
		setNAME(result,SCRATCH);
	}
	else
	{
		result = RInstall(SCRATCH,nrows*ncols);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(result,2);
		setDIM(result,1,dims[0]);
		setDIM(result,2,dims[1]);
	}
	if(isVector(result))
	{
		setDIM(result,2,0);
		setNDIMS(result,1);
	}
	if(reverse)
	{
		iop |= REVERSE;
	}
	qauto(DATAPTR(coefs),DATAPTR(x),DATAPTR(result),
		  *scratch,*scratch + nrows,ncoefs,nrows,ncols,
		  limits[0]-1,limits[1]-1,iop);

	emptyTrash();

	return (result);
	
  usage:
	myprint("ERROR: ");
  usage1:
	sprintf(OUTSTR,
			"usage is %s(%s,x[,reverse:T][,limits:cat(i1,i2),start:x1])",
			FUNCNAME, (iop <= 2) ? "phi" : "theta");

  errorExit:
	putOUTSTR();

	emptyTrash();
	Removesymbol(result);

	return (0);
} /*movavg()*/

		
