/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Pvalsub
#else
#define __SEG__ Pvalsub
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"

#define USEBETAI   /* use cumulative beta to compute cumulative binomial */
#define USEGAMMABASE   /* use cumulative gamma to compute cumulative Poisson */

/*
  This C function returns the value  P( X <= x )
  where X is Beta with parameters  a & b
*/
extern double   lgamma();

double          Cbet(x, a, b)
double          x, a, b;
{
	double          cdf;

	cdf = (x <= 0.0) ? 0 :(( x >= 1 ) ? 1.0 : betai(x,a,b)) ;

	return (cdf);
}

double          Qbeta(p, a, b)
double          p, a, b;
{
	double          ppbeta();
	long            ifault;
	return (ppbeta(p, a, b, &ifault));
}


/*
  This C function returns the value  P( X <= x )
  where X is Binomial with parameters n & P
  Calling program should check parameters, not done here
*/

#ifndef USEBETAI
double          Cbin(ix, n, P)
long            ix, n;
double          P;

{
	long            i, ix0;
	double          prob = 1.0, sum = 0.0, px0;
	double          x = (double) ix, x0;
	double          di, dipl1, dnmni, dnmnipl1, dn = (double) n;
	
	ix0 = (long) (x0 = ceil(dn * P));
	prob = exp(lgamma(dn + 1.0) + x0*log(P) + (dn - x0)*log(1.0 - P) -
			   lgamma(x0 + 1.0) - lgamma(dn - x0 + 1.0));
	px0 = prob;

	dipl1 = (double) ix0;
	dnmni = dn - dipl1 + 1.0;
	if (ix < ix0)
	{
		for (i = ix0 - 1; i >= 0; i--)
		{
			prob *= (dipl1-- / P) * (1.0 - P) / dnmni++;
			if(i <= ix)
			{
				sum += prob;
			}
		}
	}
	else
	{
		sum = px0;
		for (i = ix0 - 1; i >= 0; i--)
		{
			prob *= (dipl1-- / P) * (1.0 - P) / dnmni++;
			sum += prob;
		}
		prob = px0;
		di = (double) (ix0 + 1);
		dnmnipl1 = dn - di + 1.0;
		for (i = x0 + 1; i <= ix; i++)
		{
			prob *= ((P / di++) / (1.0 - P)) * dnmnipl1--;
			sum += prob;
		}
	}

	return (sum);
}
#else /*USEBETAI*/
double          Cbin(ix, n, p)
long            ix, n;
double          p;
{
	double          prob;
	
	if(ix < 0)
	{
		prob = 0.0;
	}
	else if(ix >= n)
	{
		prob = 1.0;
	}
	else
	{
		prob = betai(1.0 - p, (double) (n - ix), (double) (ix + 1));
	}
	
	return (prob);
}

#endif /*USEBETAI*/

/*
  This C function returns the value  P( X <= x )
  where X is chi-square with f degrees of freedom

  NOTE: This function calls gammabase
        which can be found in gammabas.c
  Changed to accept non-integral d.f. by kb
*/

double          Cchi(x, f)
double          x;
double          f;
{
	double          cdf;

	if (f <= 0)
	{
		myerrorout("The chi-square distribution must have positive df");
		return (-1);
	}

	if (x <= 0.0)
	{
		return (0.0);
	}

	f *= .5;
	x *= .5;
	gammabase(&x, &f, &cdf);
	return (cdf);

}

double          Qchi(p, f)
double          p, f;
{
	long            ifault;
	double          ppgamma();

	return (2. * ppgamma(p, f / 2., &ifault));

}

/*
  This C function returns the value  P( X <= x )
  where X is Gamma with parameters a & b.
  The likelihood of the gamma is
                (a-1)    (-b*x)
      L(a,b) = x     *exp
*/

double          Cgam(x, a, b)
double          x, a, b;
{
	double          y, cdf;

	if (a <= 0.0 || b <= 0.0)
	{
		myerrorout("Both parameters for the gamma distribution must be positive");
		return (-1);
	}

	y = x / b;
	gammabase(&y, &a, &cdf);

	return (cdf);

}

double          Qgam(p, a, b)
double          p, a, b;
{
	double          ppgamma();
	long            ifault;
	return (b * ppgamma(p, a, &ifault));
}

/*
  This C function returns the value  P( X <= x )
  where X is Hypergeometric; i.e, X counts the
  number of marked balls drawn from an urn containing
  M balls (N of which are marked) when the size
  of the draw is K.
*/

double          Chyp(x, M, N, K)
long            x, M, N, K;

{
	long            i, x0;
	double          prob = 1.0, sum = 0.0, px0;

	if (M < 0 || N < 0 || K < 0)
	{
		myerrorout("All hypergeometric parameters must be non-negative");
		return (-1);
	}
	if (M < N)
	{
		myerrorout("In the hypergeometric, the number of marked balls");
		myerrorout("  may not exceed the total number of balls");
		return (-1);
	}
	if (M < K)
	{
		myerrorout("In the hypergeometric, the number of drawn balls");
		myerrorout("  may not exceed the total number of balls");
		return (-1);
	}

	x0 = K * N / M + 1;
	for (i = 0; i < x0; i++)
	{
		prob = prob * (N - i) / (M - i) * (K - i) / (x0 - i);
	}
	for (i = 0; i < K - x0; i++)
	{
		prob = prob * (M - N - i) / (M - x0 - i);
	}
	px0 = prob;

	if (x < x0)
	{
		for (i = x0 - 1; i > x; i--)
		{
			prob *= (i + 1.0) / (N - i) * (M - N - K + i + 1.0) / (K - i);
		}
		for (i = x; i >= 0; i--)
		{
			prob *= (i + 1.0) / (N - i) * (M - N - K + i + 1.0) / (K - i);
			sum += prob;
		}
	}
	else
	{
		sum = px0;
		for (i = x0 - 1; i >= 0; i--)
		{
			prob *= (i + 1.0) / (N - i) * (M - N - K + i + 1.0) / (K - i);
			sum += prob;
		}
		prob = px0;
		for (i = x0 + 1; i <= x; i++)
		{
			prob *= (N - i + 1.0) / i * (K - i + 1.0) / (M - N - K + i);
			sum += prob;
		}
	}

	return (sum);
}

/*
  This C function returns the value  P( X <= x )
  where X is a standard Normal random variable
*/


double          Cnor(x)
double          x;
{
	double          cdf;
	normbase(&x, &cdf);
	return (cdf);
}

double          Qnor(p)
double          p;
{
	double          ppnd();
	long            ifault;
	return (ppnd(p, &ifault));
}


/*
  This C function returns the value  P( X <= x )
  where X is Poisson with parameter l.
*/

#ifndef USEGAMMABASE
double          Cpoi(x, l)
long            x;
double          l;

{
	long            i, x0;
	double          k, px0, prob = 1.0, sum = 0.0;

	if (l > 225)
	{
		myerrorout("The poisson parameter may not exceed 225");
		return (-1);
	}

	if(l <= 0.)
	{
		return ((x >= 0) ? 1.0 : 0.0);
	}
	
	x0 = ceil(l);
	k = exp(-l / x0);
	for (i = 0; i < x0; i++)
	{
		prob *= k / (x0 - i) * l;
	}
	px0 = prob;

	if (x < x0)
	{
		for (i = x0 - 1; i > x; i--)
		{
			prob *= (i + 1) / l;
		}
		for (i = x; i >= 0; i--)
		{
			prob *= (i + 1) / l;
			sum += prob;
		}
	}
	else
	{
		sum = px0;
		for (i = x0 - 1; i >= 0; i--)
		{
			prob *= (i + 1) / l;
			sum += prob;
		}
		prob = px0;
		for (i = x0 + 1; i <= x; i++)
		{
			prob *= l / i;
			sum += prob;
		}
	}
	return (sum);

}
#else /*USEGAMMABASE*/
double          Cpoi(x, l)
long            x;
double          l;

{
	double      prob;
	
	if(x < 0)
	{
		prob = 1.0;
	}
	else if(l <= 0)
	{
		prob = 0.0;
	}
	else
	{
		double	y = l;
		double 	alpha = x + 1;
		gammabase(&y, &alpha, &prob);
	}
	return (1.0 - prob);
}

#endif /*USEGAMMABASE*/

/*
  This C function returns the value  P( X <= x )
  where X is Snedecor's F with n1 d.f. in the
  numerator and n2 d.f in the denominator.

  NOTE: This function calls Cbet() above which in turn calls
  betabase in betabase.c
*/

double          Csne(x, n1, n2)
double          x, n1, n2;
{
	double          a, b, y, cdf;
	double          Cbet();

	if (n1 <= 0.0 || n2 <= 0.0)
	{
		myerrorout("Both df's in the F distribution must be positive");
		return (-1);
	}

	a = n2 / 2.0;
	b = n1 / 2.0;

	y = n2 / (n2 + n1 * x);

	cdf = Cbet(y, a, b);

	return (1.0 - cdf);

}

double          Qsne(p, n1, n2)
double          p, n1, n2;
{
	double          x, y, a, b, ppbeta();
	long            ifault;
	a = n2 / 2.;
	b = n1 / 2.;

	y = ppbeta(1. - p, a, b, &ifault);
	x = (n2 / y - n2) / n1;
	return (x);
}


/*
  This C function returns the value  P( X <= x )
  where X is noncentral F with n1 d.f. in the
  numerator and n2 d.f in the denominator and noncentrality
  parameter lam = n1*(E[MSnum]/E[MSdenom] - 1)


  NOTE: This function calls betanc() in betabase.c
  If the summation in betanc fails to converge, it returns betanc()
  + the number of terms.
*/

double          noncentF(x, lam, n1, n2)
double          x, lam, n1, n2;
{
	double          cdf;
	long            niter = 0;
	WHERE("noncentF");
	
	if(x <= 0)
	{
		cdf = 0.0;
	}
	else
	{
		x = n1*x/(n2 + n1*x);
		cdf = betanc(x,n1/2, n2/2, lam, &niter);
		if(niter)
		{
			if(cdf >= 1.0)
			{
				cdf = 1-1e-10;
			}
			cdf += (double) niter;
		}
	}
	
	return (cdf);
}
	
double          noncentBeta(x, lam, a, b)
double          x, lam, a, b;
{
	double          cdf;
	long            niter = 0;
	WHERE("noncentB");
	
	cdf = (x <= 0.0) ? 0.0 : ((x >= 1.0) ? 1.0 : betanc(x,a, b, lam, &niter));
	if(niter)
	{
		if(cdf >= 1.0)
		{
			cdf = 1-1e-10;
		}
		cdf += (double) niter;
	}
	return (cdf);
}
	
/*
  This C function returns the value  P( X <= x )
  where X is a Student's t with n degrees of freedom

  NOTE: This function calls studentbase which can be found in studentb.c
*/

#include <stdio.h>
#include <math.h>

double          Cstu(x, n)
double          x, n;
{
	double          cdf;

	if(n == floor(n))
	{
		studentbase(&x, &n, &cdf);
	}
	else
	{
		double          a, b, y;
		long            gia, gib;
		a = b = .5*n;
		y = .5*(1.0 + sqrt(1.0 - n/(n + x*x)));

		betabase(&y, &a, &b, &gia, &gib, &cdf);

		if(x < 0)
		{
			cdf = 1.0 - cdf;
		}
	}		
	return (cdf);

}

double          Qstu(p, n)
double          p, n;
{
	long            ifault;
	double          ppstudent(), t;

	if(n == floor(n))
	{
		t = ppstudent(p, n, &ifault);
	}
	else
	{
		t = ppbeta(p ,n/2.0,n/2.0,&ifault);
		t = sqrt(.25*n/(t*(1.0-t)) - n);
		if(p <= 0.5)
		{
			t = -t;
		}
	}
	return (t);
}

