/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Readdata
#else
#define __SEG__ Readdata
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include <ctype.h>

#ifdef MACINTOSH
#include "macIface.h"
#endif

#ifdef MPW
#include <StdLib.h>
#endif /*MPW*/

#define READDATA__
#include "matdat.h"

#ifdef CONSOLENAME
#define ISCONSOLE(NAME) (strcmp(NAME, CONSOLENAME) == 0)
#define CONSOLEOK
#else /*CONSOLENAME*/
#define ISCONSOLE(NAME)  0
#define CONSOLENAME ""
#endif /*CONSOLENAME*/

#define CHUNK 50 /*number of doubles to get space for at one time */

/*
  Usage of functions
	  matread(fileName, matName) [matName may be "" or missing]
      macroread(fileName, matName)
      inforead(fileName, matName)
	  vecread(fileName)
  On {mat, macro, info}read(), matName may be omitted, in which case
  it assumes the file starts with a matrix or macro in the right format.
  On {mat, macro, info}read(), keyword phrase 'quiet:T' suppresses echoing
  of the header lines.
  On a Macintosh, if fileName is "", the user selects the file using the
  standard file dialog box.
*/

#define IVECREAD           1
#define IMATREAD           2
#define IMACROREAD         3
#define IINFOREAD          4
#define ISTRUCREAD         5
#define IREAD              6

static char     *Opnames[] = 
{
	"vecread", "matread", "macroread", "inforead", (char *) 0
};


/*
	read next data value from *LINE, starting at *LINE + place
	next is set to position of next character to read
*/
static double nextItem(long place, long *next)
{
	char      *cplace = *LINE + place, *cnext;
	double     value;
	
	while(isspace(*cplace))
	{
		cplace++;
	}
	value = mystrtod(cplace, &cnext);
	*next = (cplace == cnext) ? place : cnext - *LINE;
	return (value);
} /*nextItem()*/

static void headerError(long errorNumber, char *matName, char *fileName)
{
	long        printLine = 0;
	
	switch(errorNumber)
	{
	  case NOTFOUND:
		sprintf(OUTSTR, "ERROR: dataset or macro %s not found", matName);
		break;
	  case BADHEADER:
		sprintf(OUTSTR, "ERROR: header for %s in wrong format", matName);
		printLine = 1;
		break;
	  case TWOFORMATS:
		sprintf(OUTSTR,
				"ERROR: two scanf formats found in header lines for %s",
				matName);
		printLine = 1;
		break;
	  case BADFORMAT:
		sprintf(OUTSTR, "ERROR: too few fields specified in format for %s",
				matName);
		printLine = 1;
		break;
	  case NODATA:
		sprintf(OUTSTR, "ERROR: endfile hit before data or macro reached");
		break;
	  case NOFIELDS:
		sprintf(OUTSTR,
				"ERROR: header line for %s specifies zero cases", matName);
		printLine = 1;
		break;
	  case NOLINES:
		sprintf(OUTSTR, "ERROR: header line specifies 0 lines in macro %s",
				matName);
		printLine = 1;
		break;
	  default:
		sprintf(OUTSTR, "ERROR: headerError %ld for %s", errorNumber, matName);
		printLine = 1;
	}
	strcat(OUTSTR, " on file ");
	myprint(OUTSTR);
	myerrorout(fileName);
	if(printLine)
	{
		myerrorout("Last line read was:");
		echoLINE();
	}
	*OUTSTR = '\0';
} /*headerError()*/

static void readError(int errorNumber, char *matName, char *fileName)
{
	switch (errorNumber)
	{
	  case NODATA:
		sprintf(OUTSTR, "ERROR: no data read from dataset %s on file ", matName);
		break;
	  case NOMACRO:
		sprintf(OUTSTR, "ERROR: no body found for macro %s on file ", matName);
		break;
	  default:
		sprintf(OUTSTR,
				"ERROR: problem %ld reading dataset or macro %s from file ",
				errorNumber, matName);
	}
	myprint(OUTSTR);
	myerrorout(fileName);
	myerrorout("Last line read was:");
	echoLINE();
	*OUTSTR = '\0';
} /*readError()*/

static Symbolhandle doReadVec(FILE *fn, char *fname, long echo)
{
	long            status = 1, gotsome = 1, spaceused, spaceleft;
	double        **data = (double **) 0;
	double          tmp;
	Symbolhandle    result = (Symbolhandle) 0;
	long            warnflag = 0;
	long            place = 0, next;
	int             c;
	WHERE("doReadVec");
			
	spaceused = 0;
	spaceleft = CHUNK;
	result = RInstall(SCRATCH, spaceleft);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	data = DATA(result);

	while (status != EOF)
	{
		if (spaceleft < 1)
		{		/* make more space */
			spaceleft = CHUNK;
			data = (double **) mygrowhandle((char **) data,
						 (spaceused + spaceleft) * sizeof(double));
			setDATA(result, data);
			if(data == (double **) 0)
			{
				goto errorExit;
			}
		}/*if (spaceleft < 1)*/

		while (spaceleft > 0 && gotsome)
		{
			tmp = nextItem(place, &next);
			if(next == place)
			{ /* no number read */
				while(isspace(c = (*LINE)[place++]))
				{ /* skip white space */
					;
				}

				if(c == '\0' && status != EOF)
				{ /* nothing left in LINE; refill it */
					status = fillLINE(fn);
					gotsome = ((*LINE)[0] != '\0');
					if(echo && gotsome)
					{
						echoLINE();
					}
					place = 0;
				}
				else if(c == '\0' || c == '!')
				{
					status = EOF;
					gotsome = 0;
				}
				else if (c == '?')
				{
					/* missing value code */
					setMissing((*data)[spaceused++]);
					spaceleft--;
				}
				else
				{
					/* bad character */
					if (!warnflag && PRINTWARNINGS)
					{
						myprint("WARNING: nonnumeric character(s) ignored on ");
						myerrorout(fname);
					}
					warnflag = 1;
				}
			} /*if(next == place)*/
			else
			{
				(*data)[spaceused++] = tmp;
				spaceleft--;
				place = next;
			} /*if(next == place){}else{}*/
		} /*while (spaceleft > 0 && gotsome)*/
	} /*while (status != EOF)*/
	
	if (spaceused > MAXCOORD)
	{
		myerrorout("ERROR: more than 32750 data values - too long");
		goto errorExit;
	}
	/* shrink data to its actual size */
	data = (double **) mygrowhandle((char **) data, spaceused * sizeof(double));
	setDATA(result, data);
	if(data == (double **) 0)
	{
		goto errorExit;
	}
	setNDIMS(result, 1);
	setDIM(result, 1, spaceused);
	setNCLASS(result, -1);

	return (result);

  errorExit:
  	Removesymbol(result);

	return (0);
} /*doReadVec()*/

static Symbolhandle doReads(FILE *fp, char *matName, char *fileName,
							long op, long verbose, long echo)
{
	Symbolhandle     result = (Symbolhandle) 0;
	char             format[MAXFMT + 1];
	char            *fmt = format, **string;
	double          *y;
	double           tmp, missValue;
	long             nFields, nSkip, byRows;
	long             i, size, nitems, length = 0;
	long             dims[20];
	long             type;
	long             reply, readHeader();
	long             printHeader = verbose;
	long             badLogical = 0;
	long             objectType = 0;
	long             status = 1;
	WHERE("doReads");
	
	*OUTSTR = '\0';
	setMissing(missValue);
	reply = readHeader(fp, dims, &nFields, &fmt, &nSkip, &byRows, &type,
					   &missValue, matName,
					   (op != IINFOREAD) ? printHeader : -1);
	
	if(reply != NOERROR)
	{
		if(*matName == '\0')
		{
			matName = "first data set";
		}
		headerError(reply, matName, fileName);
		goto errorExit;
	}

	if(op != IINFOREAD)
	{
		if(nFields > MAXN)
		{
			sprintf(OUTSTR,
					"ERROR: number of items specified by header is %ld > %ld",
					nFields, (long) MAXN);
			goto errorExit;
		}

		if(op == IMATREAD && type == MACRO ||
		   op == IMACROREAD && type != MACRO)
		{
			sprintf(OUTSTR, "WARNING: found %s instead of %s",
					(op == IMATREAD) ? "macro" : "data set",
					(op == IMATREAD) ? "data set" : "macro");
			putOUTSTR();
		}
	
		if(type == REAL || type == LOGIC)
		{
			if(dims[0] == 1)
			{
				dims[2] = 1;
			}
		
			if(dims[1] == 0)
			{
				myerrorout("WARNING: 0 lines of data in data set");
				return (NULLSYMBOL);
			}
		
			if(*fmt == '\0')
			{
				fmt = DefaultFormat;
			}
		
			nitems = 1;
			for(i = 1;i <= dims[0];i++)
			{
				nitems *= dims[i];
			}

			result = RInstall(SCRATCH, nitems);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			setNDIMS(result, dims[0]);
			for(i = 1;i <= dims[0];i++)
			{
				setDIM(result, i, dims[i]);
			}
			setTYPE(result, type);
			size = symbolSize(result);
	/* on non-error, returns number of items actually read, if <= expected */
			reply = readFile(fp, dims, nFields, fmt, nSkip, byRows,
							 result, echo);
			if(reply < 0)
			{
				readError(reply, matName, fileName);
				goto errorExit;
			}
			else if(reply != size)
			{
				sprintf(OUTSTR,
						"WARNING: data set incomplete; %ld items set to MISSING",
						size - reply);
				putOUTSTR();
			}
			y = DATAPTR(result);
			for(i = 0;i < size;i++)
			{
				tmp = y[i];
				if(tmp == missValue)
				{
					setMissing(y[i]);
				}
				else if(type == LOGIC && !isMissing(tmp))
				{
				
					y[i] = (double) (tmp != 0.0);
					if(tmp != 0.0 && tmp != 1.0)
					{
						badLogical = 1;
					}
				}
			} /*for(i = 0;i < size;i++)*/
		} /*if(type == REAL || type == LOGIC)*/
		else if(type == MACRO)
		{
			if(dims[1] == 0)
			{
				myerrorout("WARNING: 0 lines in macro on file");
				return (NULLSYMBOL);
			}
			result = Install(SCRATCH, type);
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			if((reply = readMacro(fp, dims[1], result, echo)) <= 0)
			{
				goto errorExit;
			}
			setNDIMS(result, 1);
			setDIM(result, 1, 1);
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: %s does not know how to read type %s object",
					FUNCNAME, typeName(type));
			goto errorExit;
		}

		if(badLogical)
		{
			myerrorout("WARNING: non 0 or 1 value(s) found in LOGICAL variable read; set to T");
		}
	} /*if(op != IINFOREAD)*/
	else
	{ /* inforead()*/
		if(verbose)
		{
			(*LINE)[strlen(*LINE)-1] = '\0'; /* trim off '\n' */
			myerrorout(*LINE); /* use myerrorout to make sure line is counted */
		}
		result = Install(SCRATCH, CHAR);
		if(result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(result, 1);
		setDIM(result, 1, 1);
		while(status != EOF &&
			  (status = fillLINE(fp), (*LINE)[0] == COMMENTSTART))
		{
			if((*LINE)[1] != '"')
			{
				size = strlen(*LINE + 1);
				if(length == 0)
				{
					setSTRING(result, string = mygethandle(size + 1));
				}
				else
				{
					setSTRING(result,
							  string = mygrowhandle(STRING(result),
													length + size + 1));
				}
				if(string == (char **) 0)
				{
					goto errorExit;
				}
				strcpy(*string + length, *LINE + 1);
				length += size;
			}
		}
		if(length == 0)
		{
			setSTRING(result, string = mygethandle(1));
			if(string == (char **) 0)
			{
				goto errorExit;
			}
			**string = '\0';
		}
	}  /*if(op != IINFOREAD){}else{}*/
	
	return (result);

  errorExit:
	putOUTSTR();

	Removesymbol(result);

	return (0);
}


Symbolhandle    readdata(list)
Symbolhandle    list;
/*
   Functions vecread(), matread(), macroread(), inforead()

  Routines to read data and macros in from file.
  'Read' has been replaced by synonym 'vecread'

  Usage:
	vecread(fileName)
	matread(fileName [,setName], [,quiet:T])
	macroread(fileName [,setName], [,quiet:T])
	inforead(fileName [,setName], [,quiet:T])

  where fileName and setName must be CHARACTER scalars.

  inforead() returns as a CHARACTER scalar the comment lines describing
  dataset or macro name, stripping off the leading ')' (COMMENTSTART)

  As usual on the Macintosh, fileName may be "".

  NOTE:  As of 931111, it is no longer legal on the Macintosh for fileName
  to be missing.
*/
{
	char           *fname, *setname;
	FILE           *fn = (FILE *) 0;
	Symbolhandle    arg1, arg2 = (Symbolhandle) 0, result = (Symbolhandle) 0;
	Symbolhandle    symhKey;
	long            nargs = NARGS(list), margs;
	long            echo = -1, verbose = -1, useConsole = 0;
	long            keyindex;
	char           *keyword, *whatread;
	long            op;
	WHERE("readdata");
	
	OUTSTR[0] = '\0';
	LINE = (char **) 0;
	
	for(op = 1;op <= 4;op++)
	{
		if(strcmp(FUNCNAME, Opnames[op - 1]) == 0)
		{
			break;
		}
	}

	/* Note:  As currently coded, keywords must be trailing arguments */
	keyword = "echo";
	if((keyindex = findKeyword(list, keyword, 1)) >= 0)
	{
		symhKey = COMPVALUE(list, keyindex);
		if(TYPE(symhKey) != LOGIC || !isScalar(symhKey))
		{
			sprintf(OUTSTR, "ERROR: value for '%s' must be T or F", keyword);
			goto errorExit;
		}
		echo = (DATAVALUE(symhKey, 0) != 0.0);
		setNAME(symhKey, USEDKEYWORD);
		nargs--;
	}
	keyword = "quiet";
	if((keyindex = findKeyword(list, keyword, 1)) >= 0)
	{
		symhKey = COMPVALUE(list, keyindex);
		if(TYPE(symhKey) != LOGIC || !isScalar(symhKey))
		{
			notTorF(keyword);
			goto errorExit;
		}
		verbose = (DATAVALUE(symhKey, 0) == 0.0);
		setNAME(symhKey, USEDKEYWORD);
		nargs--;
		if(verbose && op == IVECREAD)
		{
			sprintf(OUTSTR, "WARNING: keyword %s ignored for %s",
					keyword, FUNCNAME);
			putOUTSTR();
		}
	}
	margs = (op == IVECREAD) ? 1 : 2;

	if(nargs > margs)
	{
		sprintf(OUTSTR,
				"ERROR: %s may have at most %ld non-keyword arguments",
				FUNCNAME, margs);
		goto errorExit;
	}
	
	if(op == IMACROREAD)
	{
		whatread = "macro name";
	}
	else if(op == IMATREAD)
	{
		whatread = "data set name";
	}
	else if(op == IINFOREAD)
	{
		whatread = "data or macro name";
	}
	verbose = (verbose < 0) ? 1 : verbose;
	
	arg1 = COMPVALUE(list, 0);
	if(!argOK(arg1, CHAR, (nargs > 1) ? 0 : 1))
	{
		goto errorExit;
	}
#ifdef CONSOLEOK
	if (ISCONSOLE(NAME(arg1)) || 
	   TYPE(arg1) == CHAR && isScalar(arg1) && ISCONSOLE(STRINGPTR(arg1)))
	{
		useConsole = (!ISATTY || BDEPTH > 0) ? 2 : 1;
	}

#endif /*CONSOLEOK*/
	if(!useConsole && (!isScalar(arg1) || TYPE(arg1) != CHAR))
	{
		notCharOrString("file name");
		goto errorExit;
	}

	fname = (useConsole) ? CONSOLENAME : STRINGPTR(arg1);

	if(echo < 0)
	{
#ifdef MACINTOSH
		echo = useConsole;
#else /*MACINTOSH*/
		echo = (useConsole == 2);
#endif /*MACINTOSH*/
	}

	if(nargs > 1)
	{
		arg2 = COMPVALUE(list, 1);
		if(!argOK(arg2, 0, 2))
		{
			goto errorExit;
		}
		if(!isScalar(arg2) || TYPE(arg2) != CHAR)
		{
			notCharOrString(whatread);
			goto errorExit;
		}
	}
	
	if(margs == 2)
	{
		setname = (nargs == 1) ? "" : STRINGPTR(arg2);
		if(useConsole && *setname != '\0')
		{
			sprintf(OUTSTR,
					"WARNING: 2nd argument to %s ignored when reading from CONSOLE",
					FUNCNAME);
			putOUTSTR();
			setname = "";
		}
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}

#ifdef MACINTOSH
	if(ReadVolume)
	{
		SetVol(0L, ReadVolume);
	}
	if(*fname == '\0')
	{
	/* file name is "" */
		fname = macOpen("\pSpecify the data file", (Str255) 0, READIT,
					0, (OSType *) 0); /* macOpen default is 'TEXT' */
		if (fname == (char *) 0)
		{/* open cancelled */
			sprintf(OUTSTR, "WARNING: %s cancelled", FUNCNAME);
			goto errorExit;
		}
		ReadVolume = Reply.vRefNum;
	} /*if(*fname == '\0')*/
#endif /*MACINTOSH*/

	if(!useConsole)
	{		
		if((fname = expandFilename(fname)) == (char *) 0)
		{
			goto errorExit;
		}

		if(!isfilename(fname))
		{
			goto errorExit;
		}
	} /*if(!useConsole)*/	
	
	LINE = mygethandle(MAXLINE+1);
	if(LINE == (char **) 0)
	{
		goto errorExit;
	}
	(*LINE)[0] = '\0';
	fn = (FILE *) 0;

#ifdef CONSOLEOK
	if(useConsole)
	{
		if(op == IVECREAD)
		{
			sprintf(OUTSTR,
					"Type data separated by spaces and terminated by '!'");
		}
		else if(op == IMATREAD || op == IMACROREAD)
		{
			sprintf(OUTSTR, "Type header, followed by %s",
						   (strcmp(FUNCNAME, "matread") == 0) ? "data" : "macro");
		}
		else
		{ /* inforead() */
			sprintf(OUTSTR,
					"Type header, followed by comments starting with '%c'",
					COMMENTSTART);
		}
		
#ifndef MACINTOSH
		putOUTSTR();
#else /*MACINTOSH*/
		if(useConsole == 1)
		{
			myAlert(OUTSTR);
			*OUTSTR = '\0';
		}
		else
		{
			putOUTSTR();
		}
#endif /*MACINTOSH*/
	} /*if(useConsole)*/
#endif /*CONSOLEOK*/

	if(!useConsole)
	{
		fn = fopen(fname, "r");
		if (fn == NULL)
		{
			sprintf(OUTSTR, "ERROR: %s cannot open file %s", FUNCNAME, fname);
		}
		else if (myfeof(fn))
		{/* no data */
			sprintf(OUTSTR, "ERROR: file %s is empty", fname);
		}
	} /*if(!useConsole)*/
	else
	{ /* read from CONSOLE */
		fn = INPUTFILE[BDEPTH];
		fname = *INPUTFILENAMES[BDEPTH];
	} /*if(!useConsole){}else{}*/

	if(*OUTSTR)
	{
		goto errorExit;
	}

	if(op == IVECREAD)
	{
		result = doReadVec(fn, fname, echo);
	}
	else
	{
		result = doReads(fn, setname, fname, op, verbose, echo);
	}
	
#ifdef MACINTOSH
	if (UseWindows && useConsole && BDEPTH == 0)
	{
		InvalRect(&(CmdWind->portRect));
		macUpdate((WindowPtr) 0); /* update all windows */
	}
#endif /*MACINTOSH*/
	if(fn != STDIN && !useConsole)
	{
		fclose(fn);
	}
	*OUTSTR = '\0';
/* fall through */

  errorExit:
	putOUTSTR();

	mydisphandle (LINE);
	LINE = (char **) 0;

	return (result);
}

