/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Solve
#else
#define __SEG__ Solve
#endif /*MPW3*/
#endif /*MPW*/

#include "dbug.h"
#include "globals.h"

#include "blas.h"   /* prototypes and macros for names */
#include "linpack.h"   /* prototypes and macros for names */
#include <math.h>

/* 
  Routine to compute solve(a), solve(a,b), and det(a), where a is square
*/

#define IDET      0
#define ISOLVE    1
#define IINVERSE  2

#define GIPVT     0
#define GSCRATCH  1
#define NTRASH    2

Symbolhandle solve(Symbolhandle list)
{
	Symbolhandle     result = (Symbolhandle) 0;
	Symbolhandle     arg1, arg2;
	double         **scratchH;
	long           **ipvtH;
	long             op;
	long             nargs = NARGS(list);
	long             margs = 1, dim[2], nrows, ncols = 1;
	long             i, j;
	long             nitems, needed;
	long            *ipvt;
	long             job;
	double          *scratch, *a, *z, *x, *b;
	double           rcond, determ[2]; 
	WHERE("solver");
	TRASH(NTRASH,errorExit);
	
	*OUTSTR = '\0';
	
	if(strcmp(FUNCNAME,"det") == 0)
	{
		op = IDET;
	}
	else if(strcmp(FUNCNAME,"solve") == 0)
	{
		op = (nargs == 1) ? IINVERSE : ISOLVE;
		if(op == ISOLVE) 
		{
			margs = -2;
		}
	}
	

	if(nargs != labs(margs))
	{
		badNargs(FUNCNAME,margs);
		goto errorExit;
	}

	arg1 = COMPVALUE(list,0);
	if(!argOK(arg1,REAL,(nargs > 1) ? 1 : 0))
	{
		goto errorExit;
	}

	if(!isMatrix(arg1, dim) || dim[0] != dim[1])
	{
		sprintf(OUTSTR,
				"ERROR: %sargument to %s not square matrix",
				(nargs > 1) ? "1st " : "", FUNCNAME);
	}
	else if(anyMissing(arg1))
	{
		sprintf(OUTSTR,"ERROR: missing values in 1st argument to %s",
				FUNCNAME);
	}
	if(*OUTSTR)
	{
		goto errorExit;
	}
	
	nrows = dim[0];
	
	if(op == IDET)
	{
		needed = 1;
	}
	else if(op == ISOLVE)
	{
		arg2 = COMPVALUE(list,1);
		if(!argOK(arg2, REAL, 2))
		{
			goto errorExit;
		}
		if(!isMatrix(arg2,dim) || dim[0] != nrows)
		{
			sprintf(OUTSTR,
					"ERROR: 2nd argument to %s not matrix with same number of rows as 1st",
					FUNCNAME);
		}
		else if(anyMissing(arg2))
		{
			sprintf(OUTSTR,"ERROR: missing values in 2nd argument to %s",
					FUNCNAME);
		}
		if(*OUTSTR)
		{
			goto errorExit;
		}
		
		ncols = dim[1];
		needed = nrows * ncols;
	}
	else
	{
		ncols = nrows;
		needed = nrows * nrows;
	}

	result = RInstall(SCRATCH, needed);
	if(result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	if(ncols > 1)
	{
		setNDIMS(result,2);
		setDIM(result,1,nrows);
		setDIM(result,2,ncols);
	}
	
	needed = (op == IINVERSE) ? nrows : nrows * (nrows + 1);
	if(!getScratch(ipvtH,GIPVT,nrows,long) ||
	   !getScratch(scratchH,GSCRATCH,needed,double))
	{
		goto errorExit;
	}

	ipvt = *ipvtH;
	scratch = *scratchH;
	z = scratch;
	scratch += nrows;
	a = (op == IINVERSE) ? DATAPTR(result) : scratch;

	x = DATAPTR(arg1);
	
	nitems = nrows * nrows;
	for(i=0;i<nitems;i++)
	{
		a[i] = x[i];
	}
	
	DGECO(a, &nrows, &nrows, ipvt, &rcond, z);
	if(interrupted() != INTNOTSET)
	{
		goto errorExit;
	}

	if((1.0 + rcond) == 1.0)
	{
		sprintf(OUTSTR,
				"%s: argument to %s %ssingular",
				(op == IDET) ? "WARNING" : "ERROR",
				FUNCNAME,
				(rcond != 0.0) ? "apparently " : "");
		putOUTSTR();
		if(op != IDET)
		{
			goto errorExit;
		}
		else
		{
			DATAVALUE(result,0) = 0.0;
		}
	}

	if(op == IINVERSE || op == IDET && rcond != 0.0)
	{
		job = (op == IDET) ? 10 : 1;
		DGEDI(a, &nrows, &nrows, ipvt, determ, z, &job);
		if(interrupted() != INTNOTSET)
		{
			goto errorExit;
		}
		if(op == IDET)
		{
			DATAVALUE(result,0) = determ[0] * pow(10.0, determ[1]);
		}
	}
	else if(op == ISOLVE)
	{
		b = DATAPTR(arg2);
		x = DATAPTR(result);
		/* solve column by column */
		for(j=0;j<ncols;j++)
		{
			for(i=0;i<nrows;i++)
			{
				x[i] = b[i];
			}
			job = 0;
			DGESL(a, &nrows, &nrows, ipvt, x, &job);
			if(interrupted() != INTNOTSET)
			{
				goto errorExit;
			}
			x += nrows;
			b += nrows;
		}
	}
	
	emptyTrash();
	UNLOADSEG(DDOT);
	UNLOADSEG(DGECO);
	
	return(result);
	
  errorExit:
	putOUTSTR();
	emptyTrash();
	Removesymbol(result);
	UNLOADSEG(DDOT);
	UNLOADSEG(DGECO);
			
	return (0);
}
