/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#include "globals.h"
#include "dbug.h"

#ifdef BCPP
#include <alloc.h>
#define malloc farmalloc
#define realloc farrealloc
#define free farfree
#endif  /* BCPP */

#define DATAOFFSET sizeof(infoblock)

#define MEMINC(N) (CurrentMemory += (N) + DATAOFFSET,\
				   MaxMemory = (CurrentMemory > MaxMemory) ?\
				   CurrentMemory : MaxMemory)

#define MEMDEC(N) (CurrentMemory -= (N) + DATAOFFSET)


/*
  930504: Installed warehousing of handles of popular sizes.  When they
  are disposed of, the storage is not given up and thus they can be reused
  without call to malloc().

	Warehouse handles of sizes {1,2,...,NWAREHOUSES}*DOUBLESIZE in
	Warehouse[0],...,Warehouse[NWAREHOUSES-1], LISTSIZE
	in Warehouse[NWAREHOUSES], and SYMBOLSIZE in Warehouse[NWAREHOUSES+1]

  931105: Added function collectGarbage() to purge the warehouses

  950804: Implemented maximum warehouse sizes.  Currently all warehouse have
          length of MAXWAREHOUSELENGTH (settable in Makefile).  Another
          possibility would be to have an array MaxWareHouseLength[]
          initialized in myInitHandles() with differing lengths depending
          on the warehouse
*/

#ifndef NWAREHOUSES
#define NWAREHOUSES    9 /* number of basic warehouses */
#endif /*NWAREHOUSES*/

#define EXTRAWAREHOUSES 2 /* number of additional warehouses */

#ifndef MAXWAREHOUSELENGTH /* maximum number of handles in any warehouse */
#define MAXWAREHOUSELENGTH      16
#endif /*MAXWAREHOUSELENGTH*/

#define LISTWAREHOUSE   NWAREHOUSES
#define SYMBOLWAREHOUSE (NWAREHOUSES+1)

#define DOUBLESIZE    sizeof(double)
#define SYMBOLSIZE    sizeof(Symbol)

#define LISTSIZE (DEFAULTLISTLENGTH * sizeof(Symbolhandle))
#define MAXDBLWAREHOUSESIZE  (NWAREHOUSES * DOUBLESIZE)

#define POW2          3  /*log2 DOUBLESIZE*/
#define MASK          0x07 /*DOUBLESIZE - 1*/

#ifdef POW2
#define MODDOUBLE(A)  ((A) & MASK)
#define DIVDOUBLE(A)  ((A) >> POW2)
#else /*POW2*/
#define MODDOUBLE(A)  ((A) % DOUBLESIZE)
#define DIVDOUBLE(A)  ((A) / DOUBLESIZE)
#endif /*POW2*/

typedef union infoblock
{
	double         dummy; /* to force correct alignment */
	struct
	{
		char            *data; /* points to actual data */
		long             length;  /* number of bytes */
		union infoblock *next; /* points to next handle in warehouse, if any*/
	} info;
} infoblock, *infoblockptr;

static infoblockptr Warehouse[NWAREHOUSES+EXTRAWAREHOUSES];

/* The following are globals for debugging purposes; see mainpars.y */
short               WarehouseLength[NWAREHOUSES+EXTRAWAREHOUSES];
short               Nwarehouses = NWAREHOUSES+EXTRAWAREHOUSES; 

/* Maximum sizes for each warehouse, set by setWarehouseLimits()*/

static unsigned short  WarehouseLimits[NWAREHOUSES+EXTRAWAREHOUSES];

#ifdef COUNTMISSES
long                WarehouseMisses[NWAREHOUSES+EXTRAWAREHOUSES];
long                WarehouseHits[NWAREHOUSES+EXTRAWAREHOUSES];
#endif /*COUNTMISSES*/

#define maxWarehouseLength(I)  WarehouseLimits[I]

extern long         CurrentMemory, MaxMemory; /* defined in comhandl.c */

/* unix & MSDOS version of freeList */
static void    freeList(infoblockptr list)
{
	infoblockptr    link, next;
	char           *p;

	for(link = list;link != (infoblockptr) 0;link = next)
	{
		next = link->info.next;
		MEMDEC(link->info.length);
		link->info.data = (char *) 0;
		p = (char *) link;
		free((void *) p);
	}
} /*freeList()*/

static void    collectGarbage(void)
{
	register short  i;

	/* discard warehoused handles */
	for(i = 0;i < Nwarehouses;i++)
	{
		freeList(Warehouse[i]);
		Warehouse[i] = (infoblockptr) 0;
		WarehouseLength[i] = 0;
	}

} /*collectGarbage()*/

void **myFindHandle(void ** h)
{
	infoblockptr        p = (infoblockptr) h;

	if(p->info.data != ((char *) h) + DATAOFFSET)
	{
		myerrorout("WARNING: Probable memory corruption");
		h = (void **) 0;
	}
	return (h);
}

/*
   930505 version

   950805 added code to decrement WarehouseLength[whichWarehouse]
*/
void          **myNewHandle(length)
long            length;
{
	/* get handle to length bytes */
	char           *p;
	short           whichWarehouse = -1;
	infoblockptr    p1;
	WHERE("myNewHandle");
	
#ifdef MAXHANDLESIZE
	if (length > MAXHANDLESIZE)
	{
		myerrorout("ERROR: Request for too big a chunk of memory");
		return(BADHANDLE);
	}
#endif /*MAXHANDLESIZE*/

	if(length == SYMBOLSIZE)
	{
		whichWarehouse = SYMBOLWAREHOUSE;
	}
	else if(length == LISTSIZE)
	{
		whichWarehouse = LISTWAREHOUSE;
	}
	else if(MODDOUBLE(length) == 0 && length <= MAXDBLWAREHOUSESIZE)
	{
		whichWarehouse = DIVDOUBLE(length) - 1;
	}

	p1 = (whichWarehouse >= 0) ?
		Warehouse[whichWarehouse] : (infoblockptr) 0;

	if (p1 != (infoblockptr) 0)
	{ /* found warehoused handle */
#ifdef COUNTMISSES
		WarehouseHits[whichWarehouse]++;
#endif /*COUNTMISSES*/
		Warehouse[whichWarehouse] = p1->info.next;
		p1->info.data--; /* fixup pointer */
		p1->info.next = (infoblockptr) 0;
		p = p1->info.data - DATAOFFSET;
		WarehouseLength[whichWarehouse]--;
	}
	else
	{ /* no warehouse handle of right size */
#ifdef COUNTMISSES
		if (whichWarehouse >= 0)
		{
			WarehouseMisses[whichWarehouse]++;
		}
#endif /*COUNTMISSES*/
		p = (char *) malloc((size_t) length + DATAOFFSET);
		if (p != (char *) 0)
		{
			p1 = (infoblockptr) p;
			p1->info.data = p + DATAOFFSET;
			p1->info.length = length;
			p1->info.next = (infoblockptr) 0;
			MEMINC(length);
		}
		else
		{
			myerrorout("Compacting memory, please stand by");
/* free up warehoused handles and try again */
			collectGarbage();
			p = (char *) malloc((size_t) length + DATAOFFSET);
			if (p != (char *) 0)
			{
				p1 = (infoblockptr) p;
				p1->info.data = p + DATAOFFSET;
				p1->info.length = length;
				p1->info.next = (infoblockptr) 0;
				MEMINC(length);
			}
		}
	}

	return ((void **) p);
} /*myNewHandle()*/

/*
   930505 version
   950804: added code to check and increment WarehouseLength[whichWarehouse]
 */
void myDisposHandle(h)
void          **h;
{
	infoblockptr   p1 = (infoblockptr ) h;
	long           length;
	short          whichWarehouse = -1;
	WHERE("myDisposHandle");
	
/* dispose of a handle */
	if (h != (void **) 0)
	{
		length = p1->info.length;

		/* warehouse Symbols, standard lists and small REAL vectors */
		if(length == SYMBOLSIZE)
		{
			whichWarehouse = SYMBOLWAREHOUSE;
		}
		else if(length == LISTSIZE)
		{
			whichWarehouse = LISTWAREHOUSE;
		}
		else if(MODDOUBLE(length) == 0 && length <= MAXDBLWAREHOUSESIZE)
		{
			whichWarehouse = DIVDOUBLE(length) - 1;
		}

		if(whichWarehouse < 0 ||
		   WarehouseLength[whichWarehouse] >= maxWarehouseLength(whichWarehouse))
		{/* not warehousable */			
			MEMDEC(length);
			p1->info.data = (char *) 0;
			free((void *) h);
		}
		else
		{
			p1->info.data++; /* make it fail check if reused improperly */
			p1->info.next = Warehouse[whichWarehouse];
			Warehouse[whichWarehouse] = p1;			
			WarehouseLength[whichWarehouse]++;
		}
	} /*if (h != (void **) 0)*/
} /*myDisposHandle()*/

void          **mySetHandleSize(h, newn)
void          **h;
long            newn;
{
	/* make a handle bigger */
	char          **h1 = (char **) h, *h2;
	infoblockptr    p = (infoblockptr) h;
	void           *realloc();
	long            length = p->info.length;
	WHERE("mySetHandleSize");
	
#ifdef MAXHANDLESIZE
	if (newn > MAXHANDLESIZE)
	{
		myerrorout("ERROR: Request for too big a chunk of memory");
		return (BADHANDLE);
	}
#endif   /* MAXHANDLESIZE */
	if(length != newn)
	{ /* reallocate only if length is different */
#ifdef UNDEFINED__
/*
   following code uses realloc which can invalidate existing handle on failure
*/
		MEMDEC(length);
		h2 = (char *) realloc((void *) (*h1 - DATAOFFSET),
							  (size_t) newn + DATAOFFSET);
		if (h2 != (char *) 0)
		{
			p = (infoblockptr) h2;
			p->info.data = h2 + DATAOFFSET;
			p->info.length = newn;
			p->info.next = (infoblockptr) 0;
			MEMINC(newn);
		} /*if (h2 != (char *) 0)*/
		else
		{
/*
   Free up warehoused space.  However, it is unsafe to try again since
   the attempt may have destroyed h1
*/
			myerrorout("Reclaiming memory, please stand by");
			collectGarbage();
		}
#else /*UNDEFINED__*/
		/* allocate, copy, and dispose*/
		h2 = (char *) myNewHandle(newn);

		if (h2 != (char *) 0)
		{
			char    *new = ((infoblockptr) h2)->info.data;
			char    *old = *h1;
			
			length = (length < newn) ? length : newn;
			
#ifndef USEMEMCPY
			while (length-- > 0)
			{
				*new++ = *old++;
			}
#else /*USEMEMCPY*/
			memcpy(new, old, length);
#endif /*USEMEMCPY*/
			myDisposHandle(h);
		} /*if (h2 != (char *) 0)*/
#endif /*UNDEFINED__*/
	}
	else
	{
		h2 = (char *) h1;
	}
	
	return ((void **) h2);
} /*mySetHandleSize()*/


/* 930505 version */
void          **myDupHandle(h)
void          **h;
{
	/* duplicate a handle and its contents */
	register long   length = ((infoblockptr) h)->info.length;
	register char  *old, *new;
	char           *h2;
	WHERE("myDupHandle");
	
	h2 = (char *) myNewHandle(length);
	if (h2 != (char *) 0)
	{
		old = ((infoblockptr) h)->info.data;
		new = ((infoblockptr) h2)->info.data;
#ifndef USEMEMCPY
		while (length-- > 0)
		{
			*new++ = *old++;
		}
#else /*USEMEMCPY*/
		memcpy(new, old, length);
#endif /*USEMEMCPY*/
	}

	return ((void **) h2);
}

long myGetHandleSize(h)
void       **h;
{
	return (((infoblockptr) h)->info.length);
}

void myHLock(void **ch)
{
}

void myHUnlock(void **ch)
{
}

void *myNewPtr(long n)
{
	return ((void *) malloc((size_t) n));
}

void myDisposePtr(void *p)
{
	free(p);
}

/*
   Added 950809
*/

void setWarehouseLimits(long standard)
{
	int         i;
	
	for (i=0;i < Nwarehouses;i++)
	{
		WarehouseLimits[i] = standard;
#ifdef COUNTMISSES
		WarehouseHits[i] = WarehouseMisses[i] = 0;
#endif /*COUNTMISSES*/
	} /*for (i=0;i < Nwarehouses;i++)*/
	WarehouseLimits[0] = 10*standard;
	WarehouseLimits[SYMBOLWAREHOUSE] = 4*standard;
} /*setWarehouseLimits()*/

/*
   950805 added code to initialize WarehouseLength[]
*/

void myInitHandles(void)
{
	short              i;
	static short       first = 1;
	
	if(first)
	{ /* test should be unnecessary but play it safe */
		first = 0;
		for(i=0;i < Nwarehouses;i++)
		{
			Warehouse[i] = (infoblockptr) 0;
			WarehouseLength[i] = 0;
		}
		setWarehouseLimits(MAXWAREHOUSELENGTH);
	} /*if(first)*/
} /*myInitHandles()*/
