/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 3.35 or later
*(C)*
*(C)* Copyright (c) 1994 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef MPW
#ifdef MPW3
#pragma segment Utils
#else
#define __SEG__ Utils
#endif /*MPW3*/
#endif /*MPW*/

#include "globals.h"
#include "dbug.h"
#include <math.h>
#include <ctype.h>

/*
  Various utilities for working with structures.  Some were formerly in 
  utils.c
*/

/*
  doRecur2()
  Handle recursive part of Symbolhandle returning functions that accept
  two structures as arguments.
  Note that to recur on arg1, the LEFTRECUR bit of control must be set and
  to recur on arg2, the RIGHTRECUR bit must be set.  This allows doRecur2
  to be used with functions such as rational() that may have several
  arguments that are not structures as well as one that is.  The non-structure
  arguments can be combined into a temporary STRUC variable and passed
  as arg2 to doRecur2.
  Note: result is in symbol table; this ensures removal in case of an interrupt
  Function doit() must also return symbol in symbol table.

  If both args are structures, they must have same number of components
*/

Symbolhandle doRecur2(Symbolhandle (*doit) (Symbolhandle, Symbolhandle, double *, unsigned long, unsigned long *),
					  Symbolhandle arg1,Symbolhandle arg2,
					  double *params, unsigned long control,
					  unsigned long *status)
{
	Symbolhandle     result = (Symbolhandle) 0, symh;
	long            i;
	long            leftrecur = (isStruc(arg1) && (control & LEFTRECUR));
	long            rightrecur = (isStruc(arg2) && (control & RIGHTRECUR));
	WHERE("doRecur2");

	if(leftrecur || rightrecur)
	{
		long     ncomps = (leftrecur) ? NCOMPS(arg1) : NCOMPS(arg2);
		
		if(control & REUSELEFT)
		{
			result = arg1;
		}
		else if(control & REUSERIGHT)
		{
			result = arg2;
		}
		else
		{
			result = StrucInstall(SCRATCH,ncomps); /*put in symbol table */
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
			
		for(i=0;i<ncomps;i++)
		{
			symh = doRecur2(doit,
							(leftrecur) ?  COMPVALUE(arg1,i) : arg1,
							(rightrecur) ? COMPVALUE(arg2,i) : arg2,
							params,control,status);
			if(!(control & (REUSELEFT | REUSERIGHT)))
			{
				Cutsymbol(symh); /* remove component from symbol table */
			}
			
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			COMPVALUE(result,i) = symh;
			setCompName(symh,
						(leftrecur) ?
						NAME(COMPVALUE(arg1,i)) : NAME(COMPVALUE(arg2,i)));
		}
	} /*if(leftrecur || rightrecur)*/
	else
	{
		result = doit(arg1, arg2, params, control, status);
	}
	
	return (result);

  errorExit:
	Removesymbol(result);
	return (0);
} /*doRecur2()*/

/*
  doRecur1()
  Handle recursive part of functions that accept one structure as
  argument

  There is no need to set the LEFTRECUR bit in control and it is not checked,
  since it does not make sense for doRecur1 to be called with a STRUC argument
  unless recursion is desired.

  Note: result is in symbol table; this ensures removal in case of an interrupt
  Function doit() must also return symbol in symbol table.
*/

Symbolhandle doRecur1(Symbolhandle (*doit) (Symbolhandle, double *,
											unsigned long, unsigned long *),
					  Symbolhandle arg1, double *params,
					  unsigned long control,unsigned long *status)
{
	Symbolhandle    result = (Symbolhandle) 0, symh;
	long            i;
	long            leftrecur = isStruc(arg1) ;
	WHERE("doRecur1");

	if(leftrecur)
	{
		long     ncomps = NCOMPS(arg1);
		
		if(control & REUSELEFT)
		{
			result = arg1;
		}
		else
		{
			result = StrucInstall(SCRATCH,ncomps); /* in symbol table */
			if(result == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		}
		
		for(i=0;i<ncomps;i++)
		{
			symh = doRecur1(doit,COMPVALUE(arg1,i),params,control,status);
			if(!(control & REUSELEFT))
			{
				Cutsymbol(symh); /* remove component from symbol table */
			}
			
			if(symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			COMPVALUE(result,i) = symh;
			setCompName(symh,NAME(COMPVALUE(arg1,i)));
		}
	}
	else
	{
		result = doit(arg1, params, control, status);
	}
	
	return (result);

  errorExit:
	Removesymbol(result);
	return (0);
} /*doRecur1()*/

/*
  Test whether all components of a structure satisfy the isMatrix() criterion
  A null component is o.k.
*/

long strucIsMatrix(Symbolhandle arg)
{
	long        i, ncomps;
	
	if(isNull(arg))
	{
		return (1);
	}
	if(isStruc(arg))
	{
		ncomps = NCOMPS(arg);
		for(i=0;i<ncomps;i++)
		{
			if(!strucIsMatrix(COMPVALUE(arg,i)))
			{
				return (0);
			}
		}
		return (1);
	}
	return (isMatrix(arg,(long *) 0));
} /*strucIsMatrix()*/

/*
  Test whether all components of a structure satisfy the isVector() criterion
  A null component is o.k.
*/

long strucIsVector(Symbolhandle arg)
{
	long        i, ncomps;
	
	if(isNull(arg))
	{
		return (1);
	}
	if(isStruc(arg))
	{
		ncomps = NCOMPS(arg);
		for(i=0;i<ncomps;i++)
		{
			if(!strucIsVector(COMPVALUE(arg,i)))
			{
				return (0);
			}
		}
		return (1);
	} /*if(isStruc(arg))*/
	return (isVector(arg));
} /*strucIsVector()*/

/*
  Determine all the  types of components of a structure, ORing them together
  in the return value.  Ignore null components unless all components are
  null, inwhich case, return NULLTYPE1.
*/

long getStrucTypes(symh)
Symbolhandle    symh;
{
	long        type = 0, type1;
	long        i, ncomps;
	WHERE("getStrucTypes");
	
	if(isNull(symh))
	{
		type = NULLTYPE1;
	}
	else if(isStruc(symh))
	{
		ncomps = NCOMPS(symh);
		for(i=0;i<ncomps;i++)
		{
			type1 = getStrucTypes(COMPVALUE(symh,i));
			if(type1 != NULLTYPE1)
			{
				type |= type1;
			}
		}
		if(type == 0)
		{ /* all components were NULLTYPE1 */
			type = NULLTYPE1;
		}
	}
	else
	{
		type = TYPE(symh);
		if(type == REAL)
		{
			type = REALTYPE1;
		}
		else if(type == LOGIC)
		{
			type = LOGICTYPE1;
		}
		else if(type == CHAR)
		{
			type = CHARTYPE1;
		}
		else if(type == PLOTINFO)
		{
			type = PLOTINFOTYPE1;
		}
		else if(type == LIST)
		{
			type = LISTTYPE1;
		}
		else
		{
			type = UNKNOWNTYPE1;
		}
	}
	
	return (type);
} /*getStrucTypes()*/

/*
  getSingleType() scans a structure.  If all non-STRUC, non-null components
  have the same type, it returns that type.  If all components are null,
  it returns -1.  Otherwise it returns 0.
*/
long getSingleType(Symbolhandle symh)
{
	long         type = getStrucTypes(symh);
	WHERE("getSingleType");
	
	if(type == REALTYPE1)
	{
		type = REAL;
	}
	else if(type == LOGICTYPE1)
	{
		type = LOGIC;
	}
	else if(type == CHARTYPE1)
	{
		type = CHAR;
	}
	else if(type == STRUCTYPE1)
	{
		type = STRUC;
	}
	else if(type == PLOTINFOTYPE1)
	{
		type = PLOTINFO;
	}
	else if(type == LISTTYPE1)
	{
		type = LIST;
	}
	else if(type == NULLTYPE1)
	{
		type = -1;
	}
	else
	{
		type = 0;
	}
	
	return (type);
} /*getSingleType()*/

/*
   Scan a structure for the presence of null (null handle or 0 length)
   symbols.  If any are found, return 1, else return 0
*/
long strucAnyNull(arg)
Symbolhandle   arg;
{
	long       i, ncomps;
	
	if(isNull(arg))
	{
		return (1);
	}
	if(isStruc(arg))
	{
		ncomps = NCOMPS(arg);
		for(i=0;i<ncomps;i++)
		{
			if(strucAnyNull(COMPVALUE(arg,i)))
			{
				return (1);
			}
		}
	}
	return (0);
} /*strucAnyNull()*/

/*
   Compute sum of symbolSize(comp) for all non STRUC components of argument
*/

long strucSymbolSize(arg)
Symbolhandle      arg;
{
	long       size = 0;
	long       i, ncomps;
	int        type;
	
	if(!isNull(arg))
	{
		type = TYPE(arg);
		if(type == STRUC || type == LIST)
		{
			ncomps = NCOMPS(arg);
			for(i=0;i<ncomps;i++)
			{
				size += strucSymbolSize(COMPVALUE(arg,i));
			}
		}
		else
		{
			size = symbolSize(arg);
		}
	} /*if(!isNull(arg))*/
	return (size);
} /*strucSymbolSize()*/

/* 
   Compute number of bytes in all non-STRUC components of a structure
   Currently, if a component has type PLOTINFO, a full count of bytes is
   not given.  This does not matter at present (931208) since strucSize()
   is used only when all the components of its argument have types
   REAL, LOGIC or CHAR.
*/

long strucSize(arg)
Symbolhandle     arg;
{
	long        i, ncomps;
	long        size = 0;
	int         type;
	WHERE("strucSize");
	
	if(!isNull(arg))
	{
		type = TYPE(arg);
		if(type == STRUC || type == LIST)
		{
			ncomps = NCOMPS(arg);
			for(i=0;i<ncomps;i++)
			{
				size += strucSize(COMPVALUE(arg,i));
			}
		} /*if(type == STRUC || type == LIST)*/
		else
		{
			switch (TYPE(arg))
			{
			  case LOGIC:
			  case REAL:
				TMPHANDLE = (char **) DATA(arg);
				break;
			  case CHAR:
			  case MACRO:
				TMPHANDLE = STRING(arg);
				break;
			  case PLOTINFO:
				TMPHANDLE = (char **) GRAPH(arg);
				break;
			}
			size = myhandlelength(TMPHANDLE);
		} /*if(type == STRUC || type == LIST){}else{}*/
	} /*if(!isNull(arg))*/
	return (size);
} /*strucSize()*/
	
/*
   strucAnyMissing() returns 1 if its argument contains missing values, or
   any of its components do.
*/

long strucAnyMissing(Symbolhandle arg)
{
	long        ncomps, i;
	
	if(!isNull(arg) && TYPE(arg) == STRUC)
	{
		ncomps = NCOMPS(arg);
		for(i=0;i<ncomps;i++)
		{
			if(strucAnyMissing(COMPVALUE(arg,i)))
			{
				return (1);
			}
		}
		return (0);
	}
	else
	{
		return (anyMissing(arg));
	}
} /*strucAnyMissing()*/

	
/*
  dimcmp() returns 0 if and only if the dimensions of arg1 and arg2 match,
  with the level of strictness required specified by control.
  
  op = (control & OPMASK) specifies form of check on dimensions

  op == 0 no check on dimensions
  op == STRICT exact match of dimensions
  op == NEARSTRICT exact match except for extra trailing dimensions of 1
  op == MATMULT compatible for %*%
  op == TRMATMULT compatible for %c%
  op == MATMULTTR compatible for %C%
  Otherwise (binary op token), compatible for binary elementwise operation

  Two null components always match.
*/

long dimcmp(arg1, arg2, control)
Symbolhandle arg1, arg2;
unsigned long     control;
{
	long     ndims1, ndims2, lastDim;
	long     dim1[MAXDIMS], dim2[MAXDIMS];
	long     i, maxNdims, minNdims;
	int      op = control & OPMASK;
	long     null1 = isNull(arg1), null2 = isNull(arg2);
	WHERE("dimcmp");
	
	if(op == 0)
	{
		return (0);
	}
	
	if(null1 || null2)
	{ /* o.k. if both are null */
		return ((!null1 || !null2) ? BADDIMS : 0);
	}
	
	ndims1 = NDIMS(arg1);
	ndims2 = NDIMS(arg2);
	switch(op)
	{
	  case STRICT:
		/* extra trailing dimensions of 1 not allowed */
		if(ndims1 != ndims2)
		{
			return (BADDIMS);
		}
		for(i=1;i<=ndims1;i++)
		{
			if(DIMVAL(arg1,i) != DIMVAL(arg2,i))
			{
				return (BADDIMS);
			}
		} /*for(i=1;i<=ndims1;i++)*/
		break;
		
	  case NEARSTRICT:
		/* extra trailing dimensions of 1 are allowed */
		minNdims = (ndims1 < ndims2) ? ndims1 : ndims2;
		maxNdims = (ndims1 > ndims2) ? ndims1 : ndims2;
		for(i=1;i<=maxNdims;i++)
		{
			if(i <= minNdims && DIMVAL(arg1,i) != DIMVAL(arg2,i) ||
			   i > ndims1 && DIMVAL(arg2,i) > 1 ||
			   i > ndims2 && DIMVAL(arg1,i) > 1)
			{
				return(BADDIMS);
			}
		} /*for(i=1;i<=maxNdims;i++)*/
		break;

	  case MATMULT:
	  case TRMATMULT:
	  case MATMULTTR:
		/* check for compatibility for matrix multiplication */
		if(!isMatrix(arg1,dim1) || !isMatrix(arg2,dim2))
		{
			myerrorout("ERROR: matrix multiplication with non-matrices");
			return (BADDIMS);
		}
		if(anyMissing(arg1) || anyMissing(arg2))
		{
			return (HASMISSING);
		}
		
		if(op == MATMULT && dim1[1] != dim2[0] ||
		   op == TRMATMULT && dim1[0] != dim2[0] ||
		   op == MATMULTTR && dim1[1] != dim2[1])
		{
			sprintf(OUTSTR,
				"ERROR: Dimension mismatch in matrix multiply: %ld by %ld %s %ld by %ld.  ",
				dim1[0], dim1[1], opName(op) , dim2[0], dim2[1]
			);
			putOUTSTR();

			return (BADDIMS);
		}

		break;
		
	  default:
		/* must be elementwise operation */
		if(!isScalar(arg1) && !isScalar(arg2))
		{ /* only mismatch possible is when neither are scalars */
			maxNdims = (ndims1 > ndims2) ? ndims1 : ndims2;
			for(i=0;i<maxNdims;i++)
			{ /* copy dimensions and pad with 1's if necessary */
				if(i<ndims1)
				{
					dim1[i] = DIMVAL(arg1,i+1);
				}
				else
				{
					dim1[i] = 1;
				}
				if(i < ndims2)
				{
					dim2[i] = DIMVAL(arg2,i+1);
				}
				else
				{
					dim2[i] = 1;
				}
			} /* for(i=0;i<maxNdims;i++)*/
			lastDim = maxNdims-1;
			if(dim1[0] != dim2[0] && dim1[0] != 1 && dim2[0] != 1 ||
			   lastDim > 0 && dim1[lastDim] != dim2[lastDim] &&
			   dim1[lastDim] != 1 && dim2[lastDim] != 1)
			{  /* first and last dimensions > 1 must match */
				return (BADDIMS);
			}
			for(i=1;i<lastDim;i++)
			{/* middle dimensions must match exactly */
				if(dim1[i] != dim2[i])
				{
					return (BADDIMS);
				}
			} /*for(i=1;i<lastDim;i++)*/
		} /*if(!isScalar(arg1) && !isScalar(arg2))*/
		
		break;
	} /*switch(op)*/
	return (0);
} /*dimcmp()*/

/*
  treecmp() returns zero if and only if str1 and str2 have same tree structure,
  specifically
  (1)  both are null, or
  (2)  both are non-structures of the same type, or
  (3)  both are structures with the same number of components, and
       corresponding components have the same tree structure
  Argument control is of the form leftrecur+rightrecur+op
  leftrecur == 0 specifies left is not structure & no left recurrence
      should occur
  rightrecur == 0 specifies right is not structure & no right recurrence
      should occur
  op specifies form of check on dimensions
  op == 0 no check on dimensions
  op == STRICT exact match of dimensions
  op == NEARSTRICT exact match except for extra trailing dimensions of 1
  op == MATMULT compatible for %*%
  op == TRMATMULT compatible for %c%
  op == MATMULTTR compatible for %C%
  Otherwise, compatible for binary elementwise operation
  If there is a mismatch in tree structure, treecmp returns BADSTRUC
  If there is a mismatch in dimensions, treecmp returns BADDIMS
*/

long treecmp(arg1, arg2, control)
Symbolhandle     arg1, arg2;
unsigned long    control;
{
	Symbolhandle left, right;
	long         icomp, ncomps1 = -1, ncomps2 = -1, ncomps;
	unsigned long leftrecur = control & LEFTRECUR;
	unsigned long rightrecur = control & RIGHTRECUR;
	long         null1 = isNull(arg1), null2 = isNull(arg2);
	long         op;
	long         reply;
	WHERE("treecmp");
	
	op = control & OPMASK;
	if(null1 || null2)
	{ /* o.k. if both are null */
		return ((!null1 || !null2) ? BADSTRUC : 0);
	}

	if(leftrecur && rightrecur && isStruc(arg1) && TYPE(arg1) != TYPE(arg2))
	{
		return (BADSTRUC);
	}
	else if(!leftrecur && isStruc(arg1) || !rightrecur && isStruc(arg2))
	{
		return (BADSTRUC);
	}
	leftrecur = leftrecur && isStruc(arg1);
	rightrecur = rightrecur && isStruc(arg2);
	
	if(leftrecur || rightrecur)
	{
		if(leftrecur)
		{
			ncomps1 = NCOMPS(arg1);
		}
		if(rightrecur)
		{
			ncomps2 = NCOMPS(arg2);
		}
		if(leftrecur && rightrecur && ncomps1 != ncomps2)
		{
			return (BADSTRUC);
		}
		ncomps = (leftrecur) ? ncomps1 : ncomps2;
		for(icomp=0;icomp<ncomps;icomp++)
		{
			left = (leftrecur) ? COMPVALUE(arg1,icomp) : arg1;
			right = (rightrecur) ? COMPVALUE(arg2,icomp) : arg2;
			
			if((reply = treecmp(left,right, control)) != 0)
			{
				return (reply);
			}
		}/*for(icomp=0;icomp<ncomps;icomp++)*/
		return (0);
	} /*if(leftrecur || rightrecur)*/

	return ((op != 0) ? dimcmp(arg1,arg2,control) : 0);
} /*treecmp()*/

